"""
A class for configuring or automatically discovering settings to be used with PrefectDbtRunner.
"""

import contextlib
import tempfile
from pathlib import Path
from typing import Any, Generator

import yaml
from dbt_common.events.base_types import EventLevel
from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

from prefect.settings import get_current_settings
from prefect.utilities.asyncutils import run_coro_as_sync
from prefect.utilities.templating import (
    resolve_block_document_references,
    resolve_variables,
)
from prefect_dbt.utilities import find_profiles_dir, replace_with_env_var_call


class PrefectDbtSettings(BaseSettings):
    """
    dbt settings that directly affect the PrefectDbtRunner.
    These settings will be collected automatically from their corresponding 'DBT_'-prefixed environment variables.
    If a setting is not set in the environment or in the fields of this class, the default value will be used.

    All other dbt settings should be used as normal, e.g. in the dbt_project.yml file, env vars, or kwargs to `invoke()`.
    """

    model_config = SettingsConfigDict(env_prefix="DBT_")

    profiles_dir: Path = Field(
        default_factory=find_profiles_dir,
        description="The directory containing the dbt profiles.yml file.",
    )
    project_dir: Path = Field(
        default_factory=Path.cwd,
        description="The directory containing the dbt project.",
    )
    log_level: EventLevel = Field(
        default_factory=lambda: EventLevel(
            get_current_settings().logging.level.lower()
        ),
        description="The log level of the dbt CLI. Uses Prefect's logging level if not set.",
    )
    target_path: Path = Field(
        default=Path("target"),
        description="The path to the dbt target directory (relative to project_dir).",
    )

    def load_profiles_yml(self) -> dict[str, Any]:
        """
        Load and parse the profiles.yml file.

        Returns:
            Dict containing the parsed profiles.yml contents

        Raises:
            ValueError: If profiles.yml is not found
        """
        profiles_path = self.profiles_dir / "profiles.yml"
        if not profiles_path.exists():
            raise ValueError(f"No profiles.yml found at {profiles_path}")

        with open(profiles_path, "r") as f:
            return yaml.safe_load(f)

    @contextlib.contextmanager
    def resolve_profiles_yml(self) -> Generator[str, None, None]:
        """
        Context manager that creates a temporary directory with a resolved profiles.yml file.

        Args:
            include_profiles: Whether to include the resolved profiles.yml in the yield.

        Yields:
            str: Path to temporary directory containing the resolved profiles.yml.
                Directory and contents are automatically cleaned up after context exit.

        Example:
        ```python
            with resolve_profiles_yml() as temp_dir:
                # temp_dir contains resolved profiles.yml
                # use temp_dir for dbt operations
            # temp_dir is automatically cleaned up
        ```
        """
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir_path = Path(temp_dir)
            profiles_yml: dict[str, Any] = self.load_profiles_yml()
            profiles_yml = run_coro_as_sync(
                resolve_block_document_references(
                    profiles_yml, value_transformer=replace_with_env_var_call
                )
            )
            profiles_yml = run_coro_as_sync(resolve_variables(profiles_yml))

            temp_profiles_path = temp_dir_path / "profiles.yml"
            temp_profiles_path.write_text(
                yaml.dump(profiles_yml, default_style=None, default_flow_style=False)
            )
            yield str(temp_dir_path)



import json
from unittest.mock import MagicMock

from prefect_dbt.cloud.clients import DbtCloudMetadataClient

import prefect


def test_metadata_client_query(monkeypatch):
    mock_response = {
        "data": {
            "metrics": [
                {
                    "uniqueId": "metric.tpch.total_revenue",
                    "name": "total_revenue",
                    "packageName": "tpch",
                    "tags": [],
                    "label": "Total Revenue ($)",
                    "runId": 108952046,
                    "description": "",
                    "type": "sum",
                    "sql": "net_item_sales_amount",
                    "timestamp": "order_date",
                    "timeGrains": ["day", "week", "month"],
                    "dimensions": ["status_code", "priority_code"],
                    "meta": {},
                    "resourceType": "metric",
                    "filters": [],
                    "model": {"name": "fct_orders"},
                }
            ]
        },
        "headers": {},
    }
    urlopen_mock = MagicMock()
    urlopen_mock.getcode.return_value = 200
    urlopen_mock.return_value = urlopen_mock
    urlopen_mock.read.return_value = json.dumps(mock_response).encode()
    urlopen_mock.__enter__.return_value = urlopen_mock
    monkeypatch.setattr("urllib.request.urlopen", urlopen_mock)
    dbt_cloud_metadata_client = DbtCloudMetadataClient(api_key="my_api_key")
    mock_query = """
    {
        metrics(jobId: 123) {
            uniqueId
            name
            packageName
            tags
            label
            runId
            description
            type
            sql
            timestamp
            timeGrains
            dimensions
            meta
            resourceType
            filters {
                field
                operator
                value
            }
            model {
                name
            }
        }
    }
    """
    assert dbt_cloud_metadata_client.query(mock_query) == mock_response
    mock_headers = urlopen_mock.call_args_list[0].args[0].headers
    assert mock_headers["X-dbt-partner-source"] == "prefect"
    assert mock_headers["Authorization"] == "Bearer my_api_key"
    assert mock_headers["User-agent"] == f"prefect-{prefect.__version__}"



"""
Tests for PrefectDbtSettings class and related functionality.
"""

from pathlib import Path
from typing import Any
from unittest.mock import Mock

import pytest
import yaml
from dbt_common.events.base_types import EventLevel
from prefect_dbt.core.settings import PrefectDbtSettings
from pydantic.types import SecretStr

import prefect.exceptions
from prefect.blocks.system import Secret
from prefect.variables import Variable


@pytest.fixture
def mock_find_profiles_dir(monkeypatch: pytest.MonkeyPatch) -> Path:
    """Mock find_profiles_dir to return a predictable path."""
    mock_path = Path("/mock/profiles/dir")
    monkeypatch.setattr(
        "prefect_dbt.core.settings.find_profiles_dir", Mock(return_value=mock_path)
    )
    return mock_path


@pytest.fixture
def mock_get_current_settings(monkeypatch: pytest.MonkeyPatch) -> Mock:
    """Mock get_current_settings to return predictable logging level."""
    mock_settings = Mock()
    mock_settings.logging.level = "info"
    monkeypatch.setattr(
        "prefect_dbt.core.settings.get_current_settings",
        Mock(return_value=mock_settings),
    )
    return mock_settings


@pytest.fixture
def mock_get_current_settings_error(monkeypatch: pytest.MonkeyPatch) -> Mock:
    """Mock get_current_settings to return predictable logging level."""
    mock_settings = Mock()
    mock_settings.logging.level = "error"
    monkeypatch.setattr(
        "prefect_dbt.core.settings.get_current_settings",
        Mock(return_value=mock_settings),
    )
    return mock_settings


@pytest.fixture
def temp_profiles_dir(tmp_path: Path) -> Path:
    """Create a temporary directory with a profiles.yml file."""
    profiles_dir = tmp_path / "profiles"
    profiles_dir.mkdir()
    return profiles_dir


@pytest.fixture
def sample_profiles_yml() -> dict[str, Any]:
    """Sample profiles.yml content for testing."""
    return {
        "default": {
            "targets": {
                "dev": {
                    "type": "postgres",
                    "host": "localhost",
                    "port": 5432,
                    "user": "test_user",
                    "pass": "test_pass",
                    "dbname": "test_db",
                },
                "prod": {
                    "type": "postgres",
                    "host": "prod.example.com",
                    "port": 5432,
                    "user": "prod_user",
                    "pass": "prod_pass",
                    "dbname": "prod_db",
                },
            }
        },
        "test_profile": {
            "targets": {"test": {"type": "sqlite", "path": "/tmp/test.db"}}
        },
    }


class TestPrefectDbtSettingsInitialization:
    """Test PrefectDbtSettings initialization and default values."""

    def test_initializes_with_defaults(
        self, mock_find_profiles_dir: Path, mock_get_current_settings: Mock
    ):
        """Test that settings initialize with sensible defaults."""
        settings = PrefectDbtSettings(profiles_dir=mock_find_profiles_dir)
        assert settings.profiles_dir == mock_find_profiles_dir
        assert settings.project_dir == Path.cwd()
        assert settings.target_path == Path("target")
        assert settings.log_level == EventLevel.INFO

    def test_accepts_custom_values(self, temp_profiles_dir: Path):
        """Test that settings accept and use custom values."""
        custom_project = Path("/custom/project")
        custom_target = Path("custom_target")
        custom_log_level = EventLevel.DEBUG

        settings = PrefectDbtSettings(
            profiles_dir=temp_profiles_dir,
            project_dir=custom_project,
            target_path=custom_target,
            log_level=custom_log_level,
        )

        assert settings.profiles_dir == temp_profiles_dir
        assert settings.project_dir == custom_project
        assert settings.target_path == custom_target
        assert settings.log_level == custom_log_level

    def test_uses_prefect_logging_level_when_not_set(
        self, mock_get_current_settings_error: Mock
    ):
        """Test that settings use Prefect's logging level when not explicitly set."""
        settings = PrefectDbtSettings()
        assert settings.log_level == EventLevel.ERROR

    def test_environment_variable_loading(self, monkeypatch: pytest.MonkeyPatch):
        """Test that settings load from environment variables."""
        monkeypatch.setenv("DBT_PROJECT_DIR", "/env/project")
        monkeypatch.setenv("DBT_TARGET_PATH", "env_target")
        monkeypatch.setenv("DBT_LOG_LEVEL", "debug")

        settings = PrefectDbtSettings()

        assert settings.project_dir == Path("/env/project")
        assert settings.target_path == Path("env_target")
        assert settings.log_level == EventLevel.DEBUG

    def test_environment_variables_override_defaults(
        self, monkeypatch: pytest.MonkeyPatch, temp_profiles_dir: Path
    ):
        """Test that environment variables override default values."""
        monkeypatch.setenv("DBT_PROFILES_DIR", str(temp_profiles_dir))
        monkeypatch.setenv("DBT_PROJECT_DIR", "/env/override")

        settings = PrefectDbtSettings()

        assert settings.profiles_dir == temp_profiles_dir
        assert settings.project_dir == Path("/env/override")

    def test_invalid_log_level_raises_error(self, monkeypatch: pytest.MonkeyPatch):
        """Test that invalid log level raises validation error."""
        monkeypatch.setenv("DBT_LOG_LEVEL", "invalid_level")

        with pytest.raises(
            ValueError,
            match="Input should be 'debug', 'test', 'info', 'warn' or 'error'",
        ):
            PrefectDbtSettings()


class TestPrefectDbtSettingsProfilesLoading:
    """Test profiles.yml loading functionality."""

    def test_load_profiles_yml_success(
        self, temp_profiles_dir: Path, sample_profiles_yml: dict[str, Any]
    ):
        """Test successful loading of profiles.yml file."""
        profiles_file = temp_profiles_dir / "profiles.yml"
        with open(profiles_file, "w") as f:
            yaml.dump(sample_profiles_yml, f)

        settings = PrefectDbtSettings(profiles_dir=temp_profiles_dir)
        loaded_profiles = settings.load_profiles_yml()

        assert loaded_profiles == sample_profiles_yml
        assert "default" in loaded_profiles
        assert "test_profile" in loaded_profiles
        assert loaded_profiles["default"]["targets"]["dev"]["type"] == "postgres"
        assert loaded_profiles["test_profile"]["targets"]["test"]["type"] == "sqlite"

    def test_load_profiles_yml_missing_file(self, temp_profiles_dir: Path):
        """Test that missing profiles.yml raises appropriate error."""
        settings = PrefectDbtSettings(profiles_dir=temp_profiles_dir)

        with pytest.raises(ValueError, match="No profiles.yml found"):
            settings.load_profiles_yml()

    def test_load_profiles_yml_invalid_yaml(self, temp_profiles_dir: Path):
        """Test that invalid YAML raises appropriate error."""
        profiles_file = temp_profiles_dir / "profiles.yml"
        profiles_file.write_text("invalid: yaml: content: [")

        settings = PrefectDbtSettings(profiles_dir=temp_profiles_dir)

        with pytest.raises(yaml.YAMLError):
            settings.load_profiles_yml()

    def test_load_profiles_yml_empty_file(self, temp_profiles_dir: Path):
        """Test loading empty profiles.yml file."""
        profiles_file = temp_profiles_dir / "profiles.yml"
        profiles_file.write_text("")

        settings = PrefectDbtSettings(profiles_dir=temp_profiles_dir)
        loaded_profiles = settings.load_profiles_yml()

        assert loaded_profiles is None


class TestPrefectDbtSettingsProfilesResolution:
    """Test profiles.yml resolution with templating."""

    def test_resolve_profiles_yml_creates_temp_directory(
        self, temp_profiles_dir: Path, sample_profiles_yml: dict[str, Any]
    ):
        """Test that resolve_profiles_yml creates a temporary directory."""
        profiles_file = temp_profiles_dir / "profiles.yml"
        with open(profiles_file, "w") as f:
            yaml.dump(sample_profiles_yml, f)

        settings = PrefectDbtSettings(profiles_dir=temp_profiles_dir)

        with settings.resolve_profiles_yml() as temp_dir:
            temp_dir_path = Path(temp_dir)
            assert temp_dir_path.exists()
            assert temp_dir_path.is_dir()

            # Verify profiles.yml was created in temp directory
            temp_profiles_file = temp_dir_path / "profiles.yml"
            assert temp_profiles_file.exists()

            # Verify content was copied
            with open(temp_profiles_file, "r") as f:
                content = yaml.safe_load(f)
                assert content == sample_profiles_yml

        # Verify cleanup
        assert not temp_dir_path.exists()

    async def test_resolve_profiles_yml_with_block_references(
        self, temp_profiles_dir: Path
    ):
        """Test that resolve_profiles_yml resolves Prefect Secret block references for all fields."""
        # Create actual Prefect Secret blocks
        secret_block = Secret(value=SecretStr("secret_password"))
        await secret_block.save("db-password")

        host_secret = Secret(value=SecretStr("localhost"))
        await host_secret.save("db-host")

        user_secret = Secret(value=SecretStr("dbt_user"))
        await user_secret.save("db-user")

        dbname_secret = Secret(value=SecretStr("dbt_db"))
        await dbname_secret.save("db-dbname")

        port_secret = Secret(value=SecretStr("5432"))
        await port_secret.save("db-port")

        profiles_with_blocks = {
            "default": {
                "targets": {
                    "dev": {
                        "type": "postgres",
                        "host": "{{ prefect.blocks.secret.db-host }}",
                        "password": "{{ prefect.blocks.secret.db-password }}",
                        "user": "{{ prefect.blocks.secret.db-user }}",
                        "dbname": "{{ prefect.blocks.secret.db-dbname }}",
                        "port": "{{ prefect.blocks.secret.db-port }}",
                    }
                }
            }
        }

        profiles_file = temp_profiles_dir / "profiles.yml"
        with open(profiles_file, "w") as f:
            yaml.dump(profiles_with_blocks, f)

        settings = PrefectDbtSettings(profiles_dir=temp_profiles_dir)

        with settings.resolve_profiles_yml() as temp_dir:
            resolved_profiles_file = Path(temp_dir) / "profiles.yml"
            assert resolved_profiles_file.exists()

            with open(resolved_profiles_file, "r") as f:
                resolved_profiles = yaml.safe_load(f)

            # Check that all secret block references were resolved to environment variable calls
            assert (
                resolved_profiles["default"]["targets"]["dev"]["host"]
                == "{{ env_var('PREFECT_BLOCKS_SECRET_DB_HOST') }}"
            )
            assert (
                resolved_profiles["default"]["targets"]["dev"]["password"]
                == "{{ env_var('PREFECT_BLOCKS_SECRET_DB_PASSWORD') }}"
            )
            assert (
                resolved_profiles["default"]["targets"]["dev"]["user"]
                == "{{ env_var('PREFECT_BLOCKS_SECRET_DB_USER') }}"
            )
            assert (
                resolved_profiles["default"]["targets"]["dev"]["dbname"]
                == "{{ env_var('PREFECT_BLOCKS_SECRET_DB_DBNAME') }}"
            )
            assert (
                resolved_profiles["default"]["targets"]["dev"]["port"]
                == "{{ env_var('PREFECT_BLOCKS_SECRET_DB_PORT') }}"
            )

    async def test_resolve_profiles_yml_with_variable_references(
        self, temp_profiles_dir: Path
    ):
        """Test that resolve_profiles_yml resolves Prefect variable references."""
        # Create actual Prefect variables
        await Variable.aset(name="db_host", value="prod.example.com")
        await Variable.aset(name="db_password", value="prod_password")
        await Variable.aset(name="db_user", value="prod_user")

        profiles_with_vars = {
            "default": {
                "targets": {
                    "dev": {
                        "type": "postgres",
                        "host": "{{ prefect.variables.db_host }}",
                        "password": "{{ prefect.variables.db_password }}",
                        "user": "{{ prefect.variables.db_user }}",
                        "dbname": "prod_db",
                        "port": 5432,
                    }
                }
            }
        }

        profiles_file = temp_profiles_dir / "profiles.yml"
        with open(profiles_file, "w") as f:
            yaml.dump(profiles_with_vars, f)

        settings = PrefectDbtSettings(profiles_dir=temp_profiles_dir)

        with settings.resolve_profiles_yml() as temp_dir:
            temp_profiles_file = Path(temp_dir) / "profiles.yml"
            with open(temp_profiles_file, "r") as f:
                content = yaml.safe_load(f)

            # Verify variable references were resolved
            assert content["default"]["targets"]["dev"]["host"] == "prod.example.com"
            assert content["default"]["targets"]["dev"]["password"] == "prod_password"
            assert content["default"]["targets"]["dev"]["user"] == "prod_user"
            assert content["default"]["targets"]["dev"]["dbname"] == "prod_db"
            assert content["default"]["targets"]["dev"]["port"] == 5432

    async def test_resolve_profiles_yml_with_mixed_references(
        self, temp_profiles_dir: Path
    ):
        """Test that resolve_profiles_yml resolves both block and variable references."""
        # Create actual Prefect Secret block
        mixed_secret = Secret(value=SecretStr("mixed_secret"))
        await mixed_secret.save("mixed-secret")

        # Create actual Prefect variables
        await Variable.aset(name="mixed_host", value="mixed.example.com")
        await Variable.aset(name="mixed_user", value="mixed_user")

        profiles_with_mixed = {
            "default": {
                "targets": {
                    "dev": {
                        "type": "postgres",
                        "host": "{{ prefect.variables.mixed_host }}",
                        "password": "{{ prefect.blocks.secret.mixed-secret }}",
                        "user": "{{ prefect.variables.mixed_user }}",
                        "dbname": "mixed_db",
                        "port": 5432,
                    }
                }
            }
        }

        profiles_file = temp_profiles_dir / "profiles.yml"
        with open(profiles_file, "w") as f:
            yaml.dump(profiles_with_mixed, f)

        settings = PrefectDbtSettings(profiles_dir=temp_profiles_dir)

        with settings.resolve_profiles_yml() as temp_dir:
            resolved_profiles_file = Path(temp_dir) / "profiles.yml"
            assert resolved_profiles_file.exists()

            with open(resolved_profiles_file, "r") as f:
                resolved_profiles = yaml.safe_load(f)

            # Check that both block and variable references were resolved
            assert (
                resolved_profiles["default"]["targets"]["dev"]["host"]
                == "mixed.example.com"
            )
            assert (
                resolved_profiles["default"]["targets"]["dev"]["password"]
                == "{{ env_var('PREFECT_BLOCKS_SECRET_MIXED_SECRET') }}"
            )
            assert (
                resolved_profiles["default"]["targets"]["dev"]["user"] == "mixed_user"
            )
            assert (
                resolved_profiles["default"]["targets"]["dev"]["dbname"] == "mixed_db"
            )
            assert resolved_profiles["default"]["targets"]["dev"]["port"] == 5432

    def test_resolve_profiles_yml_cleanup_on_exception(
        self, temp_profiles_dir: Path, sample_profiles_yml: dict[str, Any]
    ):
        """Test that resolve_profiles_yml cleans up on exception."""
        profiles_file = temp_profiles_dir / "profiles.yml"
        with open(profiles_file, "w") as f:
            yaml.dump(sample_profiles_yml, f)

        settings = PrefectDbtSettings(profiles_dir=temp_profiles_dir)

        temp_dir_path = None
        try:
            with settings.resolve_profiles_yml() as temp_dir:
                temp_dir_path = Path(temp_dir)
                assert temp_dir_path.exists()
                # Simulate an exception
                raise Exception("Test exception")
        except Exception:
            # Verify cleanup happened despite exception
            if temp_dir_path:
                assert not temp_dir_path.exists()

    async def test_resolve_profiles_yml_handles_missing_blocks(
        self, temp_profiles_dir: Path
    ):
        """Test that resolve_profiles_yml handles missing block references gracefully."""
        profiles_with_missing_blocks = {
            "default": {
                "targets": {
                    "dev": {
                        "type": "postgres",
                        "host": "{{ prefect.blocks.secret.nonexistent-block }}",
                        "password": "{{ prefect.blocks.secret.another-missing }}",
                    }
                }
            }
        }

        profiles_file = temp_profiles_dir / "profiles.yml"
        with open(profiles_file, "w") as f:
            yaml.dump(profiles_with_missing_blocks, f)

        settings = PrefectDbtSettings(profiles_dir=temp_profiles_dir)

        # Should raise an error when blocks don't exist
        try:
            with settings.resolve_profiles_yml():
                pass
        except Exception as exc:
            print(f"Exception type: {type(exc)}, message: {exc}")
            assert isinstance(exc, prefect.exceptions.ObjectNotFound)
        else:
            assert False, "Exception not raised for missing blocks"

    async def test_resolve_profiles_yml_handles_missing_variables(
        self, temp_profiles_dir: Path
    ):
        """Test that resolve_profiles_yml handles missing variable references gracefully."""
        profiles_with_missing_vars = {
            "default": {
                "targets": {
                    "dev": {
                        "type": "postgres",
                        "host": "{{ prefect.variables.NONEXISTENT_VAR }}",
                        "password": "{{ prefect.variables.ANOTHER_MISSING }}",
                    }
                }
            }
        }

        profiles_file = temp_profiles_dir / "profiles.yml"
        with open(profiles_file, "w") as f:
            yaml.dump(profiles_with_missing_vars, f)

        settings = PrefectDbtSettings(profiles_dir=temp_profiles_dir)

        # Test that missing variables are handled gracefully (may return None or empty string)
        with settings.resolve_profiles_yml() as temp_dir:
            temp_profiles_file = Path(temp_dir) / "profiles.yml"
            with open(temp_profiles_file, "r") as f:
                content = yaml.safe_load(f)

            # Verify that missing variables are handled (either None, empty string, or the original template string)
            host_value = content["default"]["targets"]["dev"]["host"]
            password_value = content["default"]["targets"]["dev"]["password"]

            # The behavior depends on how Prefect handles missing variables
            # It could be None, empty string, or the original template string
            assert host_value in [None, "", "{{ prefect.variables.NONEXISTENT_VAR }}"]
            assert password_value in [
                None,
                "",
                "{{ prefect.variables.ANOTHER_MISSING }}",
            ]



import json
from unittest.mock import MagicMock

import pytest
from prefect_dbt.cli.configs import (
    BigQueryTargetConfigs,
    PostgresTargetConfigs,
    SnowflakeTargetConfigs,
)
from prefect_dbt.cli.configs.base import GlobalConfigs, TargetConfigs
from prefect_dbt.cli.credentials import DbtCliProfile
from prefect_dbt.cloud.credentials import DbtCloudCredentials
from prefect_gcp import GcpCredentials
from prefect_snowflake import SnowflakeConnector, SnowflakeCredentials
from prefect_sqlalchemy import (
    ConnectionComponents,
    SqlAlchemyConnector,
    SyncDriver,
)

from prefect.settings import PREFECT_API_SERVICES_TRIGGERS_ENABLED, temporary_settings
from prefect.testing.utilities import prefect_test_harness


@pytest.fixture
def dbt_cloud_credentials():
    return DbtCloudCredentials(api_key="my_api_key", account_id=123456789)


@pytest.fixture(scope="session", autouse=True)
def prefect_db():
    """
    Sets up test harness for temporary DB during test runs.
    """
    with temporary_settings({PREFECT_API_SERVICES_TRIGGERS_ENABLED: False}):
        with prefect_test_harness():
            yield


@pytest.fixture(autouse=True)
def google_auth_mock(monkeypatch):
    """
    Mocks out the google.auth module.
    """
    google_auth_default_mock = MagicMock()
    monkeypatch.setattr("google.auth.default", google_auth_default_mock)


@pytest.fixture
def dbt_cli_profile():
    target_configs = TargetConfigs(
        type="snowflake",
        schema="my_schema",
        threads=4,
        extras=dict(
            account="account",
            user="user",
            password="password",
            role="role",
            database="database",
            warehouse="warehouse",
            client_session_keep_alive=False,
            query_tag="query_tag",
        ),
    )
    global_configs = GlobalConfigs(
        send_anonymous_usage_stats=False,
        use_colors=True,
        partial_parse=False,
        printer_width=88,
        write_json=True,
        warn_error=False,
        log_format="json",
        debug=True,
        version_check=True,
        fail_fast=True,
        use_experimental_parser=True,
        static_parser=False,
    )
    return DbtCliProfile(
        name="jaffle_shop",
        target="dev",
        target_configs=target_configs,
        global_configs=global_configs,
    )


@pytest.fixture
def dbt_cli_profile_bare() -> DbtCliProfile:
    target_configs = TargetConfigs(
        type="custom", schema="my_schema", extras={"account": "fake"}
    )
    return DbtCliProfile(
        name="prefecto",
        target="testing",
        target_configs=target_configs,
    )


@pytest.fixture()
def service_account_info_dict(monkeypatch):
    monkeypatch.setattr(
        "google.auth.crypt._cryptography_rsa.serialization.load_pem_private_key",
        lambda *args, **kwargs: args[0],
    )
    _service_account_info = {
        "project_id": "service_project",
        "token_uri": "my-token-uri",
        "client_email": "my-client-email",
        "private_key": "my-private-key",
    }
    return _service_account_info


@pytest.fixture()
def service_account_file(monkeypatch, tmp_path, service_account_info_dict):
    monkeypatch.setattr(
        "google.auth.crypt._cryptography_rsa.serialization.load_pem_private_key",
        lambda *args, **kwargs: args[0],
    )
    _service_account_file = tmp_path / "gcp.json"
    with open(_service_account_file, "w") as f:
        json.dump(service_account_info_dict, f)
    return _service_account_file


@pytest.fixture
def google_auth(monkeypatch):
    google_auth_mock = MagicMock(name="google_auth")
    default_credentials_mock = MagicMock(
        name="default_credentials",
        quota_project_id="my_project",
    )
    google_auth_mock.default.side_effect = lambda *args, **kwargs: (
        default_credentials_mock,
        None,
    )
    monkeypatch.setattr("google.auth", google_auth_mock)
    return google_auth_mock


@pytest.fixture
def snowflake_target_configs():
    credentials = SnowflakeCredentials(
        user="user",
        password="password",
        account="account.region.aws",
        role="role",
    )
    connector = SnowflakeConnector(
        schema="public",
        database="database",
        warehouse="warehouse",
        credentials=credentials,
    )
    target_configs = SnowflakeTargetConfigs(connector=connector)
    return target_configs


@pytest.fixture
def postgres_target_configs():
    credentials = ConnectionComponents(
        driver=SyncDriver.POSTGRESQL_PSYCOPG2,
        username="prefect",
        password="prefect_password",
        database="postgres",
        host="host",
        port=8080,
    )
    target_configs = PostgresTargetConfigs(schema="my_schema", credentials=credentials)
    return target_configs


def sqlalchemy_target_configs():
    credentials = SqlAlchemyConnector(
        connection_info=ConnectionComponents(
            driver=SyncDriver.POSTGRESQL_PSYCOPG2,
            database="postgres",
            username="prefect",
            password="prefect_password",
        )
    )
    target_configs = PostgresTargetConfigs(schema="my_schema", credentials=credentials)
    return target_configs


@pytest.fixture
def bigquery_target_configs(service_account_info_dict):
    credentials = GcpCredentials(service_account_info=service_account_info_dict)
    target_configs = BigQueryTargetConfigs(credentials=credentials, schema="my_schema")
    return target_configs


@pytest.fixture
def class_target_configs():
    target_configs = TargetConfigs(type="type", schema="my_schema", threads=4)
    return target_configs


@pytest.fixture
def dict_target_configs():
    target_configs = dict(type="type", schema="my_schema", threads=4)
    return target_configs



from . import _version

from .core import PrefectDbtSettings, PrefectDbtRunner
from .cloud import DbtCloudCredentials, DbtCloudJob

# Define the mapping of CLI-related attributes to their import locations
_public_api: dict[str, tuple[str, str]] = {
    "DbtCliProfile": ("prefect_dbt", "cli"),
    "GlobalConfigs": ("prefect_dbt", "cli"),
    "MissingExtrasRequireError": ("prefect_dbt", "cli"),
    "TargetConfigs": ("prefect_dbt", "cli"),
    "DbtCoreOperation": ("prefect_dbt", "cli"),
    "SnowflakeTargetConfigs": ("prefect_dbt", "cli.configs.snowflake"),
    "BigQueryTargetConfigs": ("prefect_dbt", "cli.configs.bigquery"),
    "PostgresTargetConfigs": ("prefect_dbt", "cli.configs.postgres"),
}

# Declare API for type-checkers
__all__ = [
    "__version__",
    "PrefectDbtSettings",
    "PrefectDbtRunner",
    "DbtCloudCredentials",
    "DbtCloudJob",
]


def __getattr__(attr_name: str):
    if attr_name in _public_api:
        package, module = _public_api[attr_name]
        try:
            import importlib

            mod = importlib.import_module(f".{module}", package=package)
            result = getattr(mod, attr_name)
            return result
        except ImportError:
            if "configs" in module:  # For the database-specific configs
                return None
            raise
    raise AttributeError(f"module '{__name__}' has no attribute '{attr_name}'")


__version__ = _version.__version__



class DbtCloudException(Exception):
    """Base class for dbt Cloud exceptions"""


class DbtCloudGetRunFailed(DbtCloudException):
    """Raised when unable to retrieve dbt Cloud run"""


class DbtCloudListRunArtifactsFailed(DbtCloudException):
    """Raised when unable to list dbt Cloud run artifacts"""


class DbtCloudGetRunArtifactFailed(DbtCloudException):
    """Raised when unable to get a dbt Cloud run artifact"""


class DbtCloudJobRunFailed(DbtCloudException):
    """Raised when a triggered job run fails"""


class DbtCloudJobRunCancelled(DbtCloudException):
    """Raised when a triggered job run is cancelled"""


class DbtCloudJobRunTimedOut(DbtCloudException):
    """
    Raised when a triggered job run does not complete in the configured max
    wait seconds
    """


class DbtCloudJobRunTriggerFailed(DbtCloudException):
    """Raised when a dbt Cloud job trigger fails."""


class DbtCloudGetJobFailed(DbtCloudException):
    """Raised when unable to retrieve dbt Cloud job."""


class DbtCloudJobRunIncomplete(DbtCloudException):
    """Raised when a triggered job run is not complete."""


class DbtCloudCreateJobFailed(DbtCloudException):
    """Raised when unable to create a dbt Cloud job."""


class DbtCloudDeleteJobFailed(DbtCloudException):
    """Raised when unable to delete a dbt Cloud job."""



import pytest
import respx
from httpx import Response
from prefect_dbt.cloud.runs import (
    DbtCloudGetRunArtifactFailed,
    DbtCloudGetRunFailed,
    DbtCloudListRunArtifactsFailed,
    get_dbt_cloud_run_artifact,
    get_dbt_cloud_run_info,
    list_dbt_cloud_run_artifacts,
)

from prefect import flow


class TestGetDbtCloudRunInfo:
    async def test_get_dbt_cloud_run_info(self, dbt_cloud_credentials):
        with respx.mock(using="httpx") as respx_mock:
            respx_mock.get(
                "https://cloud.getdbt.com/api/v2/accounts/123456789/runs/12/",
                headers={"Authorization": "Bearer my_api_key"},
            ).mock(return_value=Response(200, json={"data": {"id": 10000}}))

            response = await get_dbt_cloud_run_info.fn(
                dbt_cloud_credentials=dbt_cloud_credentials,
                run_id=12,
            )

            assert response == {"id": 10000}

    async def test_get_nonexistent_run(self, dbt_cloud_credentials):
        with respx.mock(using="httpx") as respx_mock:
            respx_mock.get(
                "https://cloud.getdbt.com/api/v2/accounts/123456789/runs/12/",
                headers={"Authorization": "Bearer my_api_key"},
            ).mock(
                return_value=Response(
                    404, json={"status": {"user_message": "Not found!"}}
                )
            )
            with pytest.raises(DbtCloudGetRunFailed, match="Not found!"):
                await get_dbt_cloud_run_info.fn(
                    dbt_cloud_credentials=dbt_cloud_credentials,
                    run_id=12,
                )

    async def test_get_dbt_cloud_run_info_returns_fresh_data_on_each_call(
        self, dbt_cloud_credentials
    ):
        """Test that get_dbt_cloud_run_info returns fresh data on each call.

        This test verifies that the NO_CACHE policy is working correctly by
        ensuring that repeated calls with the same inputs return different
        results when the API returns different responses.
        """
        call_count = 0

        def side_effect(request):
            nonlocal call_count
            call_count += 1
            if call_count == 1:
                return Response(200, json={"data": {"id": 12, "status": 3}})
            else:
                return Response(200, json={"data": {"id": 12, "status": 10}})

        @flow
        async def test_flow():
            with respx.mock(using="httpx") as respx_mock:
                respx_mock.get(
                    "https://cloud.getdbt.com/api/v2/accounts/123456789/runs/12/",
                    headers={"Authorization": "Bearer my_api_key"},
                ).mock(side_effect=side_effect)

                first_response = await get_dbt_cloud_run_info(
                    dbt_cloud_credentials=dbt_cloud_credentials,
                    run_id=12,
                )
                second_response = await get_dbt_cloud_run_info(
                    dbt_cloud_credentials=dbt_cloud_credentials,
                    run_id=12,
                )
                return first_response, second_response

        first, second = await test_flow()

        assert first["status"] == 3, "First call should return status 3 (RUNNING)"
        assert second["status"] == 10, "Second call should return status 10 (SUCCESS)"
        assert call_count == 2, "API should be called twice, not cached"


class TestDbtCloudListRunArtifacts:
    async def test_list_artifacts_success(self, dbt_cloud_credentials):
        with respx.mock(using="httpx") as respx_mock:
            respx_mock.get(
                "https://cloud.getdbt.com/api/v2/accounts/123456789/runs/12/artifacts/",
                headers={"Authorization": "Bearer my_api_key"},
            ).mock(return_value=Response(200, json={"data": ["manifest.json"]}))

            response = await list_dbt_cloud_run_artifacts.fn(
                dbt_cloud_credentials=dbt_cloud_credentials,
                run_id=12,
            )

            assert response == ["manifest.json"]

    async def test_list_artifacts_with_step(self, dbt_cloud_credentials):
        with respx.mock(using="httpx") as respx_mock:
            respx_mock.get(
                "https://cloud.getdbt.com/api/v2/accounts/123456789/runs/12/artifacts/?step=1",  # noqa
                headers={"Authorization": "Bearer my_api_key"},
            ).mock(return_value=Response(200, json={"data": ["manifest.json"]}))

            response = await list_dbt_cloud_run_artifacts.fn(
                dbt_cloud_credentials=dbt_cloud_credentials, run_id=12, step=1
            )

            assert response == ["manifest.json"]

    async def test_list_artifacts_failure(self, dbt_cloud_credentials):
        with respx.mock(using="httpx") as respx_mock:
            respx_mock.get(
                "https://cloud.getdbt.com/api/v2/accounts/123456789/runs/12/artifacts/",
                headers={"Authorization": "Bearer my_api_key"},
            ).mock(
                return_value=Response(
                    500, json={"status": {"user_message": "This is what went wrong"}}
                )
            )
            with pytest.raises(
                DbtCloudListRunArtifactsFailed, match="This is what went wrong"
            ):
                await list_dbt_cloud_run_artifacts.fn(
                    dbt_cloud_credentials=dbt_cloud_credentials,
                    run_id=12,
                )


class TestDbtCloudGetRunArtifact:
    async def test_get_artifact_success(self, dbt_cloud_credentials):
        with respx.mock(using="httpx") as respx_mock:
            respx_mock.get(
                "https://cloud.getdbt.com/api/v2/accounts/123456789/runs/12/artifacts/manifest.json",  # noqa
                headers={"Authorization": "Bearer my_api_key"},
            ).mock(
                return_value=Response(
                    200,
                    json={
                        "metadata": {
                            "dbt_schema_version": "https://schemas.getdbt.com/dbt/catalog/v1.json",  # noqa
                            "dbt_version": "1.1.1",
                        }
                    },
                )
            )

            response = await get_dbt_cloud_run_artifact.fn(
                dbt_cloud_credentials=dbt_cloud_credentials,
                run_id=12,
                path="manifest.json",
            )

            assert response == {
                "metadata": {
                    "dbt_schema_version": "https://schemas.getdbt.com/dbt/catalog/v1.json",
                    "dbt_version": "1.1.1",
                }
            }

    async def test_get_non_json_artifact(self, dbt_cloud_credentials):
        with respx.mock(using="httpx") as respx_mock:
            respx_mock.get(
                "https://cloud.getdbt.com/api/v2/accounts/123456789/runs/12/artifacts/compiled/dbt_artifacts/models/dim_dbt__current_models.sql",  # noqa
                headers={"Authorization": "Bearer my_api_key"},
            ).mock(return_value=Response(200, text="Hi! I'm some SQL!"))

            response = await get_dbt_cloud_run_artifact.fn(
                dbt_cloud_credentials=dbt_cloud_credentials,
                run_id=12,
                path="compiled/dbt_artifacts/models/dim_dbt__current_models.sql",
            )

            assert response == "Hi! I'm some SQL!"

    async def test_get_artifact_with_step(self, dbt_cloud_credentials):
        with respx.mock(using="httpx") as respx_mock:
            respx_mock.get(
                "https://cloud.getdbt.com/api/v2/accounts/123456789/runs/12/artifacts/manifest.json?step=1",  # noqa
                headers={"Authorization": "Bearer my_api_key"},
            ).mock(
                return_value=Response(
                    200,
                    json={
                        "metadata": {
                            "dbt_schema_version": "https://schemas.getdbt.com/dbt/catalog/v1.json",  # noqa
                            "dbt_version": "1.1.1",
                        }
                    },
                )
            )

            response = await get_dbt_cloud_run_artifact.fn(
                dbt_cloud_credentials=dbt_cloud_credentials,
                run_id=12,
                path="manifest.json",
                step=1,
            )

            assert response == {
                "metadata": {
                    "dbt_schema_version": "https://schemas.getdbt.com/dbt/catalog/v1.json",
                    "dbt_version": "1.1.1",
                }
            }

    async def test_get_artifact_failure(self, dbt_cloud_credentials):
        with respx.mock(using="httpx") as respx_mock:
            respx_mock.get(
                "https://cloud.getdbt.com/api/v2/accounts/123456789/runs/12/artifacts/manifest.json",  # noqa
                headers={"Authorization": "Bearer my_api_key"},
            ).mock(
                return_value=Response(
                    500, json={"status": {"user_message": "This is what went wrong"}}
                )
            )
            with pytest.raises(
                DbtCloudGetRunArtifactFailed, match="This is what went wrong"
            ):
                await get_dbt_cloud_run_artifact.fn(
                    dbt_cloud_credentials=dbt_cloud_credentials,
                    run_id=12,
                    path="manifest.json",
                )



"""Module containing credentials for interacting with dbt Cloud"""

from typing import Union

from pydantic import Field, SecretStr
from typing_extensions import Literal

from prefect.blocks.abstract import CredentialsBlock
from prefect_dbt.cloud.clients import (
    DbtCloudAdministrativeClient,
    DbtCloudMetadataClient,
)


class DbtCloudCredentials(CredentialsBlock):
    """
    Credentials block for credential use across dbt Cloud tasks and flows.

    Attributes:
        api_key (SecretStr): API key to authenticate with the dbt Cloud
            administrative API. Refer to the [Authentication docs](
            https://docs.getdbt.com/dbt-cloud/api-v2#section/Authentication)
            for retrieving the API key.
        account_id (int): ID of dbt Cloud account with which to interact.
        domain (Optional[str]): Domain at which the dbt Cloud API is hosted.

    Examples:
        Load stored dbt Cloud credentials:
        ```python
        from prefect_dbt.cloud import DbtCloudCredentials

        dbt_cloud_credentials = DbtCloudCredentials.load("BLOCK_NAME")
        ```

        Use DbtCloudCredentials instance to trigger a job run:
        ```python
        from prefect_dbt.cloud import DbtCloudCredentials

        credentials = DbtCloudCredentials(api_key="my_api_key", account_id=123456789)

        async with dbt_cloud_credentials.get_administrative_client() as client:
            client.trigger_job_run(job_id=1)
        ```

        Load saved dbt Cloud credentials within a flow:
        ```python
        from prefect import flow

        from prefect_dbt.cloud import DbtCloudCredentials
        from prefect_dbt.cloud.jobs import trigger_dbt_cloud_job_run


        @flow
        def trigger_dbt_cloud_job_run_flow():
            credentials = DbtCloudCredentials.load("my-dbt-credentials")
            trigger_dbt_cloud_job_run(dbt_cloud_credentials=credentials, job_id=1)

        trigger_dbt_cloud_job_run_flow()
        ```
    """

    _block_type_name = "dbt Cloud Credentials"
    _logo_url = "https://images.ctfassets.net/gm98wzqotmnx/5zE9lxfzBHjw3tnEup4wWL/9a001902ed43a84c6c96d23b24622e19/dbt-bit_tm.png?h=250"  # noqa
    _documentation_url = "https://docs.prefect.io/integrations/prefect-dbt"  # noqa

    api_key: SecretStr = Field(
        default=...,
        title="API Key",
        description="A dbt Cloud API key to use for authentication.",
    )
    account_id: int = Field(
        default=..., title="Account ID", description="The ID of your dbt Cloud account."
    )
    domain: str = Field(
        default="cloud.getdbt.com",
        description="The base domain of your dbt Cloud instance.",
    )

    def get_administrative_client(self) -> DbtCloudAdministrativeClient:
        """
        Returns a newly instantiated client for working with the dbt Cloud
        administrative API.

        Returns:
            An authenticated dbt Cloud administrative API client.
        """
        return DbtCloudAdministrativeClient(
            api_key=self.api_key.get_secret_value(),
            account_id=self.account_id,
            domain=self.domain,
        )

    def get_metadata_client(self) -> DbtCloudMetadataClient:
        """
        Returns a newly instantiated client for working with the dbt Cloud
        metadata API.

        Example:
            Sending queries via the returned metadata client:
            ```python
            from prefect_dbt import DbtCloudCredentials

            credentials_block = DbtCloudCredentials.load("test-account")
            metadata_client = credentials_block.get_metadata_client()
            query = \"\"\"
            {
                metrics(jobId: 123) {
                    uniqueId
                    name
                    packageName
                    tags
                    label
                    runId
                    description
                    type
                    sql
                    timestamp
                    timeGrains
                    dimensions
                    meta
                    resourceType
                    filters {
                        field
                        operator
                        value
                    }
                    model {
                        name
                    }
                }
            }
            \"\"\"
            metadata_client.query(query)
            # Result:
            # {
            #   "data": {
            #     "metrics": [
            #       {
            #         "uniqueId": "metric.tpch.total_revenue",
            #         "name": "total_revenue",
            #         "packageName": "tpch",
            #         "tags": [],
            #         "label": "Total Revenue ($)",
            #         "runId": 108952046,
            #         "description": "",
            #         "type": "sum",
            #         "sql": "net_item_sales_amount",
            #         "timestamp": "order_date",
            #         "timeGrains": ["day", "week", "month"],
            #         "dimensions": ["status_code", "priority_code"],
            #         "meta": {},
            #         "resourceType": "metric",
            #         "filters": [],
            #         "model": { "name": "fct_orders" }
            #       }
            #     ]
            #   }
            # }
            ```

        Returns:
            An authenticated dbt Cloud metadata API client.
        """
        return DbtCloudMetadataClient(
            api_key=self.api_key.get_secret_value(),
            domain=f"metadata.{self.domain}",
        )

    def get_client(
        self, client_type: Literal["administrative", "metadata"]
    ) -> Union[DbtCloudAdministrativeClient, DbtCloudMetadataClient]:
        """
        Returns a newly instantiated client for working with the dbt Cloud API.

        Args:
            client_type: Type of client to return. Accepts either 'administrative'
                or 'metadata'.

        Returns:
            The authenticated client of the requested type.
        """
        get_client_method = getattr(self, f"get_{client_type}_client", None)
        if get_client_method is None:
            raise ValueError(f"'{client_type}' is not a supported client type.")
        return get_client_method()



import pytest
from prefect_dbt.cli.credentials import DbtCliProfile, GlobalConfigs, TargetConfigs
from pydantic import ValidationError
from typing_extensions import Literal


@pytest.mark.parametrize("configs_type", ["dict", "model"])
def test_dbt_cli_profile_init(configs_type: Literal["dict", "model"]):
    target_configs = dict(type="snowflake", schema="schema")
    global_configs = dict(use_colors=False)
    if configs_type == "model":
        target_configs = TargetConfigs.model_validate(target_configs)
        global_configs = GlobalConfigs.model_validate(global_configs)

    dbt_cli_profile = DbtCliProfile(
        name="test_name",
        target="dev",
        target_configs=target_configs,
        global_configs=global_configs,
    )
    assert isinstance(dbt_cli_profile.target_configs, TargetConfigs)
    assert dbt_cli_profile.name == "test_name"
    assert dbt_cli_profile.target == "dev"


def test_dbt_cli_profile_init_validation_failed():
    # 6 instead of 2 now because it tries to validate each of the Union types
    with pytest.raises(ValidationError):
        DbtCliProfile(
            name="test_name", target="dev", target_configs={"extras": {"field": "abc"}}
        )


def test_dbt_cli_profile_get_profile():
    target_configs = dict(type="snowflake", schema="analysis")
    global_configs = dict(use_colors=False)
    dbt_cli_profile = DbtCliProfile(
        name="test_name",
        target="dev",
        target_configs=target_configs,
        global_configs=global_configs,
    )
    actual = dbt_cli_profile.get_profile()
    expected_target_configs = target_configs.copy()
    expected_target_configs["threads"] = 4
    expected = {
        "config": global_configs,
        "test_name": {
            "target": "dev",
            "outputs": {"dev": expected_target_configs},
        },
    }
    assert actual == expected


@pytest.mark.parametrize(
    "target_configs_request",
    [
        "snowflake_target_configs",
        "dict_target_configs",
        "class_target_configs",
    ],
)
async def test_dbt_cli_profile_save_load_roundtrip(
    target_configs_request: str, request: pytest.FixtureRequest
):
    target_configs = request.getfixturevalue(target_configs_request)
    dbt_cli_profile = DbtCliProfile(
        name="my_name",
        target="dev",
        target_configs=target_configs,
    )
    block_name = target_configs_request.replace("_", "-")
    await dbt_cli_profile.save(block_name)
    dbt_cli_profile_loaded = await DbtCliProfile.load(block_name)
    assert dbt_cli_profile_loaded.get_profile()



from pathlib import Path
from unittest.mock import patch

import pytest
from prefect_dbt.cli.configs.base import GlobalConfigs, TargetConfigs

SAMPLE_PROFILES = {
    "jaffle_shop": {
        "outputs": {
            "dev": {
                "type": "duckdb",
                "path": "jaffle_shop.duckdb",
                "schema": "main",
                "threads": 4,
            },
            "prod": {
                "type": "duckdb",
                "path": "/data/prod/jaffle_shop.duckdb",
                "schema": "main",
                "threads": 8,
            },
        },
        "target": "prod",
    },
    "other_project": {
        "outputs": {
            "dev": {
                "type": "duckdb",
                "path": "other_project.duckdb",
                "schema": "analytics",
                "threads": 4,
            }
        },
        "target": "dev",
    },
    "config": {"partial_parse": True},
}


@pytest.fixture
def mock_load_profiles():
    with patch(
        "prefect_dbt.core.settings.PrefectDbtSettings.load_profiles_yml"
    ) as mock:
        mock.return_value = SAMPLE_PROFILES
        yield mock


def test_target_configs_get_configs():
    target_configs = TargetConfigs(
        type="snowflake",
        schema="schema_input",
        threads=5,
        extras={
            "extra_input": 1,
            "null_input": None,
            "key_path": Path("path/to/key"),
        },
    )
    assert hasattr(target_configs, "_is_anonymous")
    # get_configs ignore private attrs
    assert target_configs.get_configs() == dict(
        type="snowflake",
        schema="schema_input",
        threads=5,
        extra_input=1,
        key_path=str(Path("path/to/key")),
    )


def test_target_configs_get_configs_duplicate_keys():
    with pytest.raises(ValueError, match="The keyword, schema"):
        target_configs = TargetConfigs(
            type="snowflake",
            schema="schema_input",
            threads=5,
            extras={"extra_input": 1, "schema": "something else"},
        )
        target_configs.get_configs()


def test_global_configs():
    global_configs = GlobalConfigs(log_format="json", send_anonymous_usage_stats=False)
    assert global_configs.log_format == "json"
    assert global_configs.send_anonymous_usage_stats is False


def test_from_profiles_yml_default_profile_target(mock_load_profiles):
    target_configs = TargetConfigs.from_profiles_yml()

    assert target_configs.type == "duckdb"
    assert target_configs.schema_ == "main"
    assert target_configs.threads == 8
    assert target_configs.extras == {"path": "/data/prod/jaffle_shop.duckdb"}


def test_from_profiles_yml_explicit_profile_target(mock_load_profiles):
    target_configs = TargetConfigs.from_profiles_yml(
        profile_name="other_project", target_name="dev"
    )

    assert target_configs.type == "duckdb"
    assert target_configs.schema_ == "analytics"
    assert target_configs.threads == 4
    assert target_configs.extras == {"path": "other_project.duckdb"}


def test_from_profiles_yml_invalid_profile(mock_load_profiles):
    with pytest.raises(ValueError, match="Profile invalid_profile not found"):
        TargetConfigs.from_profiles_yml(profile_name="invalid_profile")


def test_from_profiles_yml_invalid_target(mock_load_profiles):
    with pytest.raises(ValueError, match="Target invalid_target not found"):
        TargetConfigs.from_profiles_yml(
            profile_name="jaffle_shop", target_name="invalid_target"
        )


def test_from_profiles_yml_no_outputs(mock_load_profiles):
    mock_load_profiles.return_value = {"broken": {"some_other_key": {}}}
    with pytest.raises(ValueError, match="No outputs found in profile broken"):
        TargetConfigs.from_profiles_yml(profile_name="broken")


def test_from_profiles_yml_no_schema(mock_load_profiles):
    mock_load_profiles.return_value = {
        "test": {
            "outputs": {
                "dev": {
                    "type": "postgres",
                    "threads": 4,
                    # Missing schema field
                    "host": "localhost",
                }
            },
            "target": "dev",
        }
    }
    with pytest.raises(ValueError, match="No schema found"):
        TargetConfigs.from_profiles_yml(profile_name="test")


def test_from_profiles_yml_alternative_schema_keys(mock_load_profiles):
    mock_profiles = {
        "test": {
            "outputs": {
                "dev": {
                    "type": "bigquery",
                    "threads": 4,
                    "dataset": "my_dataset",  # Alternative to schema
                    "project": "my_project",
                }
            },
            "target": "dev",
        }
    }
    mock_load_profiles.return_value = mock_profiles

    target_configs = TargetConfigs.from_profiles_yml(profile_name="test")
    assert target_configs.schema_ == "my_dataset"



"""Module containing tasks and flows for interacting with dbt Cloud jobs"""

import asyncio
import shlex
import time
from json import JSONDecodeError
from typing import Any, Awaitable, Callable, Dict, List, Optional, Union

from httpx import HTTPStatusError
from pydantic import Field
from typing_extensions import Literal

from prefect import flow, task
from prefect.blocks.abstract import JobBlock, JobRun
from prefect.context import FlowRunContext
from prefect.logging import get_run_logger
from prefect.utilities.asyncutils import sync_compatible
from prefect_dbt.cloud.credentials import DbtCloudCredentials
from prefect_dbt.cloud.exceptions import (
    DbtCloudCreateJobFailed,
    DbtCloudDeleteJobFailed,
    DbtCloudGetJobFailed,
    DbtCloudGetRunArtifactFailed,
    DbtCloudGetRunFailed,
    DbtCloudJobRunCancelled,
    DbtCloudJobRunFailed,
    DbtCloudJobRunIncomplete,
    DbtCloudJobRunTimedOut,
    DbtCloudJobRunTriggerFailed,
    DbtCloudListRunArtifactsFailed,
)
from prefect_dbt.cloud.models import TriggerJobRunOptions
from prefect_dbt.cloud.runs import (
    DbtCloudJobRunStatus,
    get_dbt_cloud_run_artifact,
    get_dbt_cloud_run_info,
    list_dbt_cloud_run_artifacts,
    wait_for_dbt_cloud_job_run,
)
from prefect_dbt.cloud.utils import extract_user_message

EXE_COMMANDS = ("build", "run", "test", "seed", "snapshot")


@task(
    name="Get dbt Cloud job details",
    description="Retrieves details of a dbt Cloud job "
    "for the job with the given job_id.",
    retries=3,
    retry_delay_seconds=10,
)
async def get_dbt_cloud_job_info(
    dbt_cloud_credentials: DbtCloudCredentials,
    job_id: int,
    order_by: Optional[str] = None,
) -> Dict:
    """
    A task to retrieve information about a dbt Cloud job.

    Args:
        dbt_cloud_credentials: Credentials for authenticating with dbt Cloud.
        job_id: The ID of the job to get.

    Returns:
        The job data returned by the dbt Cloud administrative API.

    Example:
        Get status of a dbt Cloud job:
        ```python
        from prefect import flow

        from prefect_dbt.cloud import DbtCloudCredentials
        from prefect_dbt.cloud.jobs import get_job

        @flow
        def get_job_flow():
            credentials = DbtCloudCredentials(api_key="my_api_key", account_id=123456789)

            return get_job(
                dbt_cloud_credentials=credentials,
                job_id=42
            )

        get_job_flow()
        ```
    """  # noqa
    try:
        async with dbt_cloud_credentials.get_administrative_client() as client:
            response = await client.get_job(
                job_id=job_id,
                order_by=order_by,
            )
    except HTTPStatusError as ex:
        raise DbtCloudGetJobFailed(extract_user_message(ex)) from ex
    return response.json()["data"]


@task(
    name="Create dbt Cloud job",
    description="Creates a new dbt Cloud job.",
    retries=3,
    retry_delay_seconds=10,
)
async def create_dbt_cloud_job(
    dbt_cloud_credentials: DbtCloudCredentials,
    project_id: int,
    environment_id: int,
    name: str,
    execute_steps: Optional[List[str]] = None,
    **kwargs: Any,
) -> Dict:
    """
    A task to create a new dbt Cloud job.

    Args:
        dbt_cloud_credentials: Credentials for authenticating with dbt Cloud.
        project_id: The ID of the project to create the job in.
        environment_id: The ID of the environment for the job.
        name: The name of the job.
        execute_steps: List of dbt commands to execute (e.g. ["dbt run", "dbt test"]).
            Defaults to ["dbt build"].
        **kwargs: Additional job configuration options.

    Returns:
        The job data returned by the dbt Cloud administrative API.

    Example:
        Create a dbt Cloud job:
        ```python
        from prefect import flow

        from prefect_dbt.cloud import DbtCloudCredentials
        from prefect_dbt.cloud.jobs import create_dbt_cloud_job

        @flow
        def create_job_flow():
            credentials = DbtCloudCredentials(api_key="my_api_key", account_id=123456789)

            return create_dbt_cloud_job(
                dbt_cloud_credentials=credentials,
                project_id=12345,
                environment_id=67890,
                name="My New Job",
                execute_steps=["dbt run -s my_model", "dbt test -s my_model"],
            )

        create_job_flow()
        ```
    """
    logger = get_run_logger()

    logger.info(f"Creating dbt Cloud job '{name}' in project {project_id}")

    try:
        async with dbt_cloud_credentials.get_administrative_client() as client:
            response = await client.create_job(
                project_id=project_id,
                environment_id=environment_id,
                name=name,
                execute_steps=execute_steps,
                **kwargs,
            )
    except HTTPStatusError as ex:
        raise DbtCloudCreateJobFailed(extract_user_message(ex)) from ex

    job_data = response.json()["data"]
    logger.info(f"Created dbt Cloud job with ID {job_data.get('id')}")
    return job_data


@task(
    name="Delete dbt Cloud job",
    description="Deletes a dbt Cloud job.",
    retries=3,
    retry_delay_seconds=10,
)
async def delete_dbt_cloud_job(
    dbt_cloud_credentials: DbtCloudCredentials,
    job_id: int,
) -> None:
    """
    A task to delete a dbt Cloud job.

    Args:
        dbt_cloud_credentials: Credentials for authenticating with dbt Cloud.
        job_id: The ID of the job to delete.

    Example:
        Delete a dbt Cloud job:
        ```python
        from prefect import flow

        from prefect_dbt.cloud import DbtCloudCredentials
        from prefect_dbt.cloud.jobs import delete_dbt_cloud_job

        @flow
        def delete_job_flow():
            credentials = DbtCloudCredentials(api_key="my_api_key", account_id=123456789)

            delete_dbt_cloud_job(
                dbt_cloud_credentials=credentials,
                job_id=42,
            )

        delete_job_flow()
        ```
    """
    logger = get_run_logger()

    logger.info(f"Deleting dbt Cloud job with ID {job_id}")

    try:
        async with dbt_cloud_credentials.get_administrative_client() as client:
            await client.delete_job(job_id=job_id)
    except HTTPStatusError as ex:
        raise DbtCloudDeleteJobFailed(extract_user_message(ex)) from ex

    logger.info(f"Deleted dbt Cloud job with ID {job_id}")


@task(
    name="Trigger dbt Cloud job run",
    description="Triggers a dbt Cloud job run for the job "
    "with the given job_id and optional overrides.",
    retries=3,
    retry_delay_seconds=10,
)
async def trigger_dbt_cloud_job_run(
    dbt_cloud_credentials: DbtCloudCredentials,
    job_id: int,
    options: Optional[TriggerJobRunOptions] = None,
) -> Dict:
    """
    A task to trigger a dbt Cloud job run.

    Args:
        dbt_cloud_credentials: Credentials for authenticating with dbt Cloud.
        job_id: The ID of the job to trigger.
        options: An optional TriggerJobRunOptions instance to specify overrides
            for the triggered job run.

    Returns:
        The run data returned from the dbt Cloud administrative API.

    Examples:
        Trigger a dbt Cloud job run:
        ```python
        from prefect import flow

        from prefect_dbt.cloud import DbtCloudCredentials
        from prefect_dbt.cloud.jobs import trigger_dbt_cloud_job_run


        @flow
        def trigger_dbt_cloud_job_run_flow():
            credentials = DbtCloudCredentials(api_key="my_api_key", account_id=123456789)

            trigger_dbt_cloud_job_run(dbt_cloud_credentials=credentials, job_id=1)


        trigger_dbt_cloud_job_run_flow()
        ```

        Trigger a dbt Cloud job run with overrides:
        ```python
        from prefect import flow

        from prefect_dbt.cloud import DbtCloudCredentials
        from prefect_dbt.cloud.jobs import trigger_dbt_cloud_job_run
        from prefect_dbt.cloud.models import TriggerJobRunOptions


        @flow
        def trigger_dbt_cloud_job_run_flow():
            credentials = DbtCloudCredentials(api_key="my_api_key", account_id=123456789)

            trigger_dbt_cloud_job_run(
                dbt_cloud_credentials=credentials,
                job_id=1,
                options=TriggerJobRunOptions(
                    git_branch="staging",
                    schema_override="dbt_cloud_pr_123",
                    dbt_version_override="0.18.0",
                    target_name_override="staging",
                    timeout_seconds_override=3000,
                    generate_docs_override=True,
                    threads_override=8,
                    steps_override=[
                        "dbt seed",
                        "dbt run --fail-fast",
                        "dbt test --fail-fast",
                    ],
                ),
            )


        trigger_dbt_cloud_job_run()
        ```
    """  # noqa
    logger = get_run_logger()

    logger.info(f"Triggering run for job with ID {job_id}")

    try:
        async with dbt_cloud_credentials.get_administrative_client() as client:
            response = await client.trigger_job_run(job_id=job_id, options=options)
    except HTTPStatusError as ex:
        raise DbtCloudJobRunTriggerFailed(extract_user_message(ex)) from ex

    run_data = response.json()["data"]

    if "project_id" in run_data and "id" in run_data:
        logger.info(
            f"Run successfully triggered for job with ID {job_id}. "
            "You can view the status of this run at "
            f"https://{dbt_cloud_credentials.domain}/#/accounts/"
            f"{dbt_cloud_credentials.account_id}/projects/{run_data['project_id']}/"
            f"runs/{run_data['id']}/"
        )

    return run_data


@task(
    name="Get dbt Cloud job run ID",
    description="Extracts the run ID from a trigger job run API response",
)
def get_run_id(obj: Dict):
    """
    Task that extracts the run ID from a trigger job run API response,

    This task is mainly used to maintain dependency tracking between the
    `trigger_dbt_cloud_job_run` task and downstream tasks/flows that use the run ID.

    Args:
        obj: The JSON body from the trigger job run response.

    Example:
        ```python
        from prefect import flow
        from prefect_dbt.cloud import DbtCloudCredentials
        from prefect_dbt.cloud.jobs import trigger_dbt_cloud_job_run, get_run_id


        @flow
        def trigger_run_and_get_id():
            dbt_cloud_credentials=DbtCloudCredentials(
                    api_key="my_api_key",
                    account_id=123456789
                )

            triggered_run_data = trigger_dbt_cloud_job_run(
                dbt_cloud_credentials=dbt_cloud_credentials,
                job_id=job_id,
                options=trigger_job_run_options,
            )
            run_id = get_run_id(triggered_run_data)
            return run_id

        trigger_run_and_get_id()
        ```
    """
    id = obj.get("id")
    if id is None:
        raise RuntimeError("Unable to determine run ID for triggered job.")
    return id


@flow(
    name="Trigger dbt Cloud job run and wait for completion",
    description="Triggers a dbt Cloud job run and waits for the"
    "triggered run to complete.",
)
async def trigger_dbt_cloud_job_run_and_wait_for_completion(
    dbt_cloud_credentials: DbtCloudCredentials,
    job_id: int,
    trigger_job_run_options: Optional[TriggerJobRunOptions] = None,
    max_wait_seconds: int = 900,
    poll_frequency_seconds: int = 10,
    retry_filtered_models_attempts: int = 3,
) -> Dict:
    """
    Flow that triggers a job run and waits for the triggered run to complete.

    Args:
        dbt_cloud_credentials: Credentials for authenticating with dbt Cloud.
        job_id: The ID of the job to trigger.
        trigger_job_run_options: An optional TriggerJobRunOptions instance to
            specify overrides for the triggered job run.
        max_wait_seconds: Maximum number of seconds to wait for job to complete
        poll_frequency_seconds: Number of seconds to wait in between checks for
            run completion.
        retry_filtered_models_attempts: Number of times to retry models selected by `retry_status_filters`.

    Raises:
        DbtCloudJobRunCancelled: The triggered dbt Cloud job run was cancelled.
        DbtCloudJobRunFailed: The triggered dbt Cloud job run failed.
        RuntimeError: The triggered dbt Cloud job run ended in an unexpected state.

    Returns:
        The run data returned by the dbt Cloud administrative API.

    Examples:
        Trigger a dbt Cloud job and wait for completion as a stand alone flow:
        ```python
        import asyncio
        from prefect_dbt.cloud import DbtCloudCredentials
        from prefect_dbt.cloud.jobs import trigger_dbt_cloud_job_run_and_wait_for_completion

        asyncio.run(
            trigger_dbt_cloud_job_run_and_wait_for_completion(
                dbt_cloud_credentials=DbtCloudCredentials(
                    api_key="my_api_key",
                    account_id=123456789
                ),
                job_id=1
            )
        )
        ```

        Trigger a dbt Cloud job and wait for completion as a sub-flow:
        ```python
        from prefect import flow
        from prefect_dbt.cloud import DbtCloudCredentials
        from prefect_dbt.cloud.jobs import trigger_dbt_cloud_job_run_and_wait_for_completion

        @flow
        def my_flow():
            ...
            run_result = trigger_dbt_cloud_job_run_and_wait_for_completion(
                dbt_cloud_credentials=DbtCloudCredentials(
                    api_key="my_api_key",
                    account_id=123456789
                ),
                job_id=1
            )
            ...

        my_flow()
        ```

        Trigger a dbt Cloud job with overrides:
        ```python
        import asyncio
        from prefect_dbt.cloud import DbtCloudCredentials
        from prefect_dbt.cloud.jobs import trigger_dbt_cloud_job_run_and_wait_for_completion
        from prefect_dbt.cloud.models import TriggerJobRunOptions

        asyncio.run(
            trigger_dbt_cloud_job_run_and_wait_for_completion(
                dbt_cloud_credentials=DbtCloudCredentials(
                    api_key="my_api_key",
                    account_id=123456789
                ),
                job_id=1,
                trigger_job_run_options=TriggerJobRunOptions(
                    git_branch="staging",
                    schema_override="dbt_cloud_pr_123",
                    dbt_version_override="0.18.0",
                    target_name_override="staging",
                    timeout_seconds_override=3000,
                    generate_docs_override=True,
                    threads_override=8,
                    steps_override=[
                        "dbt seed",
                        "dbt run --fail-fast",
                        "dbt test --fail fast",
                    ],
                ),
            )
        )
        ```
    """  # noqa
    logger = get_run_logger()

    triggered_run_data_future = await trigger_dbt_cloud_job_run(
        dbt_cloud_credentials=dbt_cloud_credentials,
        job_id=job_id,
        options=trigger_job_run_options,
    )
    run_id = (triggered_run_data_future).get("id")
    if run_id is None:
        raise RuntimeError("Unable to determine run ID for triggered job.")

    final_run_status, run_data = await wait_for_dbt_cloud_job_run(
        run_id=run_id,
        dbt_cloud_credentials=dbt_cloud_credentials,
        max_wait_seconds=max_wait_seconds,
        poll_frequency_seconds=poll_frequency_seconds,
    )

    if final_run_status == DbtCloudJobRunStatus.SUCCESS:
        try:
            list_run_artifacts_future = await list_dbt_cloud_run_artifacts(
                dbt_cloud_credentials=dbt_cloud_credentials,
                run_id=run_id,
            )
            run_data["artifact_paths"] = list_run_artifacts_future
        except DbtCloudListRunArtifactsFailed as ex:
            logger.warning(
                "Unable to retrieve artifacts for job run with ID %s. Reason: %s",
                run_id,
                ex,
            )
        logger.info(
            "dbt Cloud job run with ID %s completed successfully!",
            run_id,
        )
        return run_data
    elif final_run_status == DbtCloudJobRunStatus.CANCELLED:
        raise DbtCloudJobRunCancelled(
            f"Triggered job run with ID {run_id} was cancelled."
        )
    elif final_run_status == DbtCloudJobRunStatus.FAILED:
        while retry_filtered_models_attempts > 0:
            logger.info(
                f"Retrying job run with ID: {run_id} "
                f"{retry_filtered_models_attempts} more times"
            )
            try:
                retry_filtered_models_attempts -= 1
                run_data = await retry_dbt_cloud_job_run_subset_and_wait_for_completion(
                    dbt_cloud_credentials=dbt_cloud_credentials,
                    run_id=run_id,
                    trigger_job_run_options=trigger_job_run_options,
                    max_wait_seconds=max_wait_seconds,
                    poll_frequency_seconds=poll_frequency_seconds,
                )
                return run_data
            except Exception:
                pass
        else:
            raise DbtCloudJobRunFailed(f"Triggered job run with ID: {run_id} failed.")
    else:
        raise RuntimeError(
            f"Triggered job run with ID: {run_id} ended with unexpected"
            f"status {final_run_status.value}."
        )


async def _build_trigger_job_run_options(
    dbt_cloud_credentials: DbtCloudCredentials,
    trigger_job_run_options: TriggerJobRunOptions,
    run_id: int,
    run_info: Dict[str, Any],
    job_info: Dict[str, Any],
):
    """
    Compiles a list of steps (commands) to retry, then either build trigger job
    run options from scratch if it does not exist, else overrides the existing.
    """
    generate_docs = job_info.get("generate_docs", False)
    generate_sources = job_info.get("generate_sources", False)

    steps_override = []
    for run_step in run_info["run_steps"]:
        status = run_step["status_humanized"].lower()
        # Skipping cloning, profile setup, and dbt deps - always the first three
        # steps in any run, and note, index starts at 1 instead of 0
        if run_step["index"] <= 3 or status == "success":
            continue
        # get dbt build from "Invoke dbt with `dbt build`"
        command = run_step["name"].partition("`")[2].partition("`")[0]

        # These steps will be re-run regardless if
        # generate_docs or generate_sources are enabled for a given job
        # so if we don't skip, it'll run twice
        freshness_in_command = (
            "dbt source snapshot-freshness" in command
            or "dbt source freshness" in command
        )
        if "dbt docs generate" in command and generate_docs:
            continue
        elif freshness_in_command and generate_sources:
            continue

        # find an executable command like `build` or `run`
        # search in a list so that there aren't false positives, like
        # `"run" in "dbt run-operation"`, which is True; we actually want
        # `"run" in ["dbt", "run-operation"]` which is False
        command_components = shlex.split(command)
        for exe_command in EXE_COMMANDS:
            if exe_command in command_components:
                break
        else:
            exe_command = ""

        is_exe_command = exe_command in EXE_COMMANDS
        is_not_success = status in ("error", "skipped", "cancelled")
        is_skipped = status == "skipped"
        if (not is_exe_command and is_not_success) or (is_exe_command and is_skipped):
            # if no matches like `run-operation`, we will be rerunning entirely
            # or if it's one of the expected commands and is skipped
            steps_override.append(command)
        else:
            # errors and failures are when we need to inspect to figure
            # out the point of failure
            try:
                run_artifact_future = await get_dbt_cloud_run_artifact.with_options(
                    retries=0, retry_delay_seconds=0
                )(
                    dbt_cloud_credentials=dbt_cloud_credentials,
                    run_id=run_id,
                    path="run_results.json",
                    step=run_step["index"],
                )
                run_artifact = run_artifact_future
            except JSONDecodeError:
                # get the run results scoped to the step which had an error
                # an error here indicates that either:
                # 1) the fail-fast flag was set, in which case
                #    the run_results.json file was never created; or
                # 2) there was a problem on dbt Cloud's side saving
                #    this artifact
                steps_override.append(command)
            else:
                # we only need to find the individual nodes for those run commands
                run_results = run_artifact["results"]
                # select nodes that were not successful
                # note "fail" here instead of "cancelled" because
                # nodes do not have a cancelled state
                run_nodes = " ".join(
                    run_result["unique_id"].split(".")[2]
                    for run_result in run_results
                    if run_result["status"] in ("error", "skipped", "fail")
                )

                select_arg = None
                if "-s" in command_components:
                    select_arg = "-s"
                elif "--select" in command_components:
                    select_arg = "--select"

                # prevent duplicate --select/-s statements
                if select_arg is not None:
                    # dbt --fail-fast run, -s, bad_mod --vars '{"env": "prod"}' to:
                    # dbt --fail-fast run -s other_mod bad_mod --vars '{"env": "prod"}'
                    command_start, select_arg, command_end = command.partition(
                        select_arg
                    )
                    modified_command = (
                        f"{command_start} {select_arg} {run_nodes} {command_end}"
                    )
                else:
                    # dbt --fail-fast, build, --vars '{"env": "prod"}' to:
                    # dbt --fail-fast build --select bad_model --vars '{"env": "prod"}'
                    dbt_global_args, exe_command, exe_args = command.partition(
                        exe_command
                    )
                    modified_command = (
                        f"{dbt_global_args} {exe_command} -s {run_nodes} {exe_args}"
                    )
                steps_override.append(modified_command)

    if trigger_job_run_options is None:
        trigger_job_run_options_override = TriggerJobRunOptions(
            steps_override=steps_override
        )
    else:
        trigger_job_run_options_override = trigger_job_run_options.copy()
        trigger_job_run_options_override.steps_override = steps_override
    return trigger_job_run_options_override


@flow(
    name="Retry subset of dbt Cloud job run and wait for completion",
    description=(
        "Retries a subset of dbt Cloud job run, filtered by select statuses, "
        "and waits for the triggered retry to complete."
    ),
)
async def retry_dbt_cloud_job_run_subset_and_wait_for_completion(
    dbt_cloud_credentials: DbtCloudCredentials,
    run_id: int,
    trigger_job_run_options: Optional[TriggerJobRunOptions] = None,
    max_wait_seconds: int = 900,
    poll_frequency_seconds: int = 10,
) -> Dict:
    """
    Flow that retrys a subset of dbt Cloud job run, filtered by select statuses,
    and waits for the triggered retry to complete.

    Args:
        dbt_cloud_credentials: Credentials for authenticating with dbt Cloud.
        trigger_job_run_options: An optional TriggerJobRunOptions instance to
            specify overrides for the triggered job run.
        max_wait_seconds: Maximum number of seconds to wait for job to complete
        poll_frequency_seconds: Number of seconds to wait in between checks for
            run completion.
        run_id: The ID of the job run to retry.

    Raises:
        ValueError: If `trigger_job_run_options.steps_override` is set by the user.

    Returns:
        The run data returned by the dbt Cloud administrative API.

    Examples:
        Retry a subset of models in a dbt Cloud job run and wait for completion:
        ```python
        from prefect import flow

        from prefect_dbt.cloud import DbtCloudCredentials
        from prefect_dbt.cloud.jobs import retry_dbt_cloud_job_run_subset_and_wait_for_completion

        @flow
        def retry_dbt_cloud_job_run_subset_and_wait_for_completion_flow():
            credentials = DbtCloudCredentials.load("MY_BLOCK_NAME")
            retry_dbt_cloud_job_run_subset_and_wait_for_completion(
                dbt_cloud_credentials=credentials,
                run_id=88640123,
            )

        retry_dbt_cloud_job_run_subset_and_wait_for_completion_flow()
        ```
    """  # noqa
    if trigger_job_run_options and trigger_job_run_options.steps_override is not None:
        raise ValueError(
            "Do not set `steps_override` in `trigger_job_run_options` "
            "because this flow will automatically set it"
        )

    run_info_future = await get_dbt_cloud_run_info(
        dbt_cloud_credentials=dbt_cloud_credentials,
        run_id=run_id,
        include_related=["run_steps"],
    )
    run_info = run_info_future

    job_id = run_info["job_id"]
    job_info_future = await get_dbt_cloud_job_info(
        dbt_cloud_credentials=dbt_cloud_credentials,
        job_id=job_id,
    )
    job_info = job_info_future

    trigger_job_run_options_override = await _build_trigger_job_run_options(
        dbt_cloud_credentials=dbt_cloud_credentials,
        trigger_job_run_options=trigger_job_run_options,
        run_id=run_id,
        run_info=run_info,
        job_info=job_info,
    )

    # to circumvent `RuntimeError: The task runner is already started!`
    flow_run_context = FlowRunContext.get()
    task_runner_type = type(flow_run_context.task_runner)

    run_data = await trigger_dbt_cloud_job_run_and_wait_for_completion.with_options(
        task_runner=task_runner_type()
    )(
        dbt_cloud_credentials=dbt_cloud_credentials,
        job_id=job_id,
        retry_filtered_models_attempts=0,
        trigger_job_run_options=trigger_job_run_options_override,
        max_wait_seconds=max_wait_seconds,
        poll_frequency_seconds=poll_frequency_seconds,
    )
    return run_data


class DbtCloudJobRun(JobRun):  # NOT A BLOCK
    """
    Class that holds the information and methods to interact
    with the resulting run of a dbt Cloud job.
    """

    def __init__(self, run_id: int, dbt_cloud_job: "DbtCloudJob"):
        self.run_id = run_id
        self._dbt_cloud_job = dbt_cloud_job
        self._dbt_cloud_credentials = dbt_cloud_job.dbt_cloud_credentials

    @property
    def _log_prefix(self):
        return f"dbt Cloud job {self._dbt_cloud_job.job_id} run {self.run_id}."

    async def _wait_until_state(
        self,
        in_final_state_fn: Awaitable[Callable],
        get_state_fn: Awaitable[Callable],
        log_state_fn: Callable = None,
        timeout_seconds: int = 60,
        interval_seconds: int = 1,
    ):
        """
        Wait until the job run reaches a specific state.

        Args:
            in_final_state_fn: An async function that accepts a run state
                and returns a boolean indicating whether the job run is
                in a final state.
            get_state_fn: An async function that returns
                the current state of the job run.
            log_state_fn: A callable that accepts a run
                state and makes it human readable.
            timeout_seconds: The maximum amount of time, in seconds, to wait
                for the job run to reach the final state.
            interval_seconds: The number of seconds to wait between checks of
                the job run's state.
        """
        start_time = time.time()
        last_state = run_state = None
        while not in_final_state_fn(run_state):
            run_state = await get_state_fn()
            if run_state != last_state:
                if self.logger is not None:
                    self.logger.info(
                        "%s has new state: %s",
                        self._log_prefix,
                        log_state_fn(run_state),
                    )
                last_state = run_state

            elapsed_time_seconds = time.time() - start_time
            if elapsed_time_seconds > timeout_seconds:
                raise DbtCloudJobRunTimedOut(
                    f"Max wait time of {timeout_seconds} seconds exceeded while waiting"
                )
            await asyncio.sleep(interval_seconds)

    @sync_compatible
    async def get_run(self) -> Dict[str, Any]:
        """
        Makes a request to the dbt Cloud API to get the run data.

        Returns:
            The run data.
        """
        try:
            dbt_cloud_credentials = self._dbt_cloud_credentials
            async with dbt_cloud_credentials.get_administrative_client() as client:
                response = await client.get_run(self.run_id)
        except HTTPStatusError as ex:
            raise DbtCloudGetRunFailed(extract_user_message(ex)) from ex
        run_data = response.json()["data"]
        return run_data

    @sync_compatible
    async def get_status_code(self) -> int:
        """
        Makes a request to the dbt Cloud API to get the run status.

        Returns:
            The run status code.
        """
        run_data = await self.get_run()
        run_status_code = run_data.get("status")
        return run_status_code

    @sync_compatible
    async def wait_for_completion(self) -> None:
        """
        Waits for the job run to reach a terminal state.
        """
        await self._wait_until_state(
            in_final_state_fn=DbtCloudJobRunStatus.is_terminal_status_code,
            get_state_fn=self.get_status_code,
            log_state_fn=DbtCloudJobRunStatus,
            timeout_seconds=self._dbt_cloud_job.timeout_seconds,
            interval_seconds=self._dbt_cloud_job.interval_seconds,
        )

    @sync_compatible
    async def fetch_result(self, step: Optional[int] = None) -> Dict[str, Any]:
        """
        Gets the results from the job run. Since the results
        may not be ready, use wait_for_completion before calling this method.

        Args:
            step: The index of the step in the run to query for artifacts. The
                first step in the run has the index 1. If the step parameter is
                omitted, then this method will return the artifacts compiled
                for the last step in the run.
        """
        run_data = await self.get_run()
        run_status = DbtCloudJobRunStatus(run_data.get("status"))
        if run_status == DbtCloudJobRunStatus.SUCCESS:
            try:
                async with (
                    self._dbt_cloud_credentials.get_administrative_client() as client
                ):  # noqa
                    response = await client.list_run_artifacts(
                        run_id=self.run_id, step=step
                    )
                run_data["artifact_paths"] = response.json()["data"]
                self.logger.info("%s completed successfully!", self._log_prefix)
            except HTTPStatusError as ex:
                raise DbtCloudListRunArtifactsFailed(extract_user_message(ex)) from ex
            return run_data
        elif run_status == DbtCloudJobRunStatus.CANCELLED:
            raise DbtCloudJobRunCancelled(f"{self._log_prefix} was cancelled.")
        elif run_status == DbtCloudJobRunStatus.FAILED:
            raise DbtCloudJobRunFailed(f"{self._log_prefix} has failed.")
        else:
            raise DbtCloudJobRunIncomplete(
                f"{self._log_prefix} is still running; "
                "use wait_for_completion() to wait until results are ready."
            )

    @sync_compatible
    async def get_run_artifacts(
        self,
        path: Literal["manifest.json", "catalog.json", "run_results.json"],
        step: Optional[int] = None,
    ) -> Union[Dict[str, Any], str]:
        """
        Get an artifact generated for a completed run.

        Args:
            path: The relative path to the run artifact.
            step: The index of the step in the run to query for artifacts. The
                first step in the run has the index 1. If the step parameter is
                omitted, then this method will return the artifacts compiled
                for the last step in the run.

        Returns:
            The contents of the requested manifest. Returns a `Dict` if the
                requested artifact is a JSON file and a `str` otherwise.
        """
        try:
            dbt_cloud_credentials = self._dbt_cloud_credentials
            async with dbt_cloud_credentials.get_administrative_client() as client:
                response = await client.get_run_artifact(
                    run_id=self.run_id, path=path, step=step
                )
        except HTTPStatusError as ex:
            raise DbtCloudGetRunArtifactFailed(extract_user_message(ex)) from ex

        if path.endswith(".json"):
            artifact_contents = response.json()
        else:
            artifact_contents = response.text
        return artifact_contents

    def _select_unsuccessful_commands(
        self,
        run_results: List[Dict[str, Any]],
        command_components: List[str],
        command: str,
        exe_command: str,
    ) -> List[str]:
        """
        Select nodes that were not successful and rebuild a command.
        """
        # note "fail" here instead of "cancelled" because
        # nodes do not have a cancelled state
        run_nodes = " ".join(
            run_result["unique_id"].split(".")[2]
            for run_result in run_results
            if run_result["status"] in ("error", "skipped", "fail")
        )

        select_arg = None
        if "-s" in command_components:
            select_arg = "-s"
        elif "--select" in command_components:
            select_arg = "--select"

        # prevent duplicate --select/-s statements
        if select_arg is not None:
            # dbt --fail-fast run, -s, bad_mod --vars '{"env": "prod"}' to:
            # dbt --fail-fast run -s other_mod bad_mod --vars '{"env": "prod"}'
            command_start, select_arg, command_end = command.partition(select_arg)
            modified_command = f"{command_start} {select_arg} {run_nodes} {command_end}"  # noqa
        else:
            # dbt --fail-fast, build, --vars '{"env": "prod"}' to:
            # dbt --fail-fast build --select bad_model --vars '{"env": "prod"}'
            dbt_global_args, exe_command, exe_args = command.partition(exe_command)
            modified_command = (
                f"{dbt_global_args} {exe_command} -s {run_nodes} {exe_args}"
            )
        return modified_command

    async def _build_trigger_job_run_options(
        self,
        job: Dict[str, Any],
        run: Dict[str, Any],
    ) -> TriggerJobRunOptions:
        """
        Compiles a list of steps (commands) to retry, then either build trigger job
        run options from scratch if it does not exist, else overrides the existing.
        """
        generate_docs = job.get("generate_docs", False)
        generate_sources = job.get("generate_sources", False)

        steps_override = []
        for run_step in run["run_steps"]:
            status = run_step["status_humanized"].lower()
            # Skipping cloning, profile setup, and dbt deps - always the first three
            # steps in any run, and note, index starts at 1 instead of 0
            if run_step["index"] <= 3 or status == "success":
                continue
            # get dbt build from "Invoke dbt with `dbt build`"
            command = run_step["name"].partition("`")[2].partition("`")[0]

            # These steps will be re-run regardless if
            # generate_docs or generate_sources are enabled for a given job
            # so if we don't skip, it'll run twice
            freshness_in_command = (
                "dbt source snapshot-freshness" in command
                or "dbt source freshness" in command
            )
            if "dbt docs generate" in command and generate_docs:
                continue
            elif freshness_in_command and generate_sources:
                continue

            # find an executable command like `build` or `run`
            # search in a list so that there aren't false positives, like
            # `"run" in "dbt run-operation"`, which is True; we actually want
            # `"run" in ["dbt", "run-operation"]` which is False
            command_components = shlex.split(command)
            for exe_command in EXE_COMMANDS:
                if exe_command in command_components:
                    break
            else:
                exe_command = ""

            is_exe_command = exe_command in EXE_COMMANDS
            is_not_success = status in ("error", "skipped", "cancelled")
            is_skipped = status == "skipped"
            if (not is_exe_command and is_not_success) or (
                is_exe_command and is_skipped
            ):
                # if no matches like `run-operation`, we will be rerunning entirely
                # or if it's one of the expected commands and is skipped
                steps_override.append(command)
            else:
                # errors and failures are when we need to inspect to figure
                # out the point of failure
                try:
                    run_artifact = await self.get_run_artifacts(
                        "run_results.json", run_step["index"]
                    )
                except JSONDecodeError:
                    # get the run results scoped to the step which had an error
                    # an error here indicates that either:
                    # 1) the fail-fast flag was set, in which case
                    #    the run_results.json file was never created; or
                    # 2) there was a problem on dbt Cloud's side saving
                    #    this artifact
                    steps_override.append(command)
                else:
                    # we only need to find the individual nodes
                    # for those run commands
                    run_results = run_artifact["results"]
                    modified_command = self._select_unsuccessful_commands(
                        run_results=run_results,
                        command_components=command_components,
                        command=command,
                        exe_command=exe_command,
                    )
                    steps_override.append(modified_command)

        if self._dbt_cloud_job.trigger_job_run_options is None:
            trigger_job_run_options_override = TriggerJobRunOptions(
                steps_override=steps_override
            )
        else:
            trigger_job_run_options_override = (
                self._dbt_cloud_job.trigger_job_run_options.copy()
            )
            trigger_job_run_options_override.steps_override = steps_override
        return trigger_job_run_options_override

    @sync_compatible
    async def retry_failed_steps(self) -> "DbtCloudJobRun":  # noqa: F821
        """
        Retries steps that did not complete successfully in a run.

        Returns:
            A representation of the dbt Cloud job run.
        """
        job = await self._dbt_cloud_job.get_job()
        run = await self.get_run()

        trigger_job_run_options_override = await self._build_trigger_job_run_options(
            job=job, run=run
        )

        num_steps = len(trigger_job_run_options_override.steps_override)
        if num_steps == 0:
            self.logger.info(f"{self._log_prefix} does not have any steps to retry.")
        else:
            self.logger.info(f"{self._log_prefix} has {num_steps} steps to retry.")
            run = await self._dbt_cloud_job.trigger(
                trigger_job_run_options=trigger_job_run_options_override,
            )
        return run


class DbtCloudJob(JobBlock):
    """
    Block that holds the information and methods to interact with a dbt Cloud job.

    Attributes:
        dbt_cloud_credentials: The credentials to use to authenticate with dbt Cloud.
        job_id: The id of the dbt Cloud job.
        timeout_seconds: The number of seconds to wait for the job to complete.
        interval_seconds:
            The number of seconds to wait between polling for job completion.
        trigger_job_run_options: The options to use when triggering a job run.

    Examples:
        Load a configured dbt Cloud job block.
        ```python
        from prefect_dbt.cloud import DbtCloudJob

        dbt_cloud_job = DbtCloudJob.load("BLOCK_NAME")
        ```

        Triggers a dbt Cloud job, waits for completion, and fetches the results.
        ```python
        from prefect import flow
        from prefect_dbt.cloud import DbtCloudCredentials, DbtCloudJob

        @flow
        def dbt_cloud_job_flow():
            dbt_cloud_credentials = DbtCloudCredentials.load("dbt-token")
            dbt_cloud_job = DbtCloudJob.load(
                dbt_cloud_credentials=dbt_cloud_credentials,
                job_id=154217
            )
            dbt_cloud_job_run = dbt_cloud_job.trigger()
            dbt_cloud_job_run.wait_for_completion()
            dbt_cloud_job_run.fetch_result()
            return dbt_cloud_job_run

        dbt_cloud_job_flow()
        ```
    """

    _block_type_name = "dbt Cloud Job"
    _logo_url = "https://images.ctfassets.net/gm98wzqotmnx/5zE9lxfzBHjw3tnEup4wWL/9a001902ed43a84c6c96d23b24622e19/dbt-bit_tm.png?h=250"  # noqa
    _documentation_url = "https://docs.prefect.io/integrations/prefect-dbt"  # noqa

    dbt_cloud_credentials: DbtCloudCredentials = Field(
        default=...,
        description="The dbt Cloud credentials to use to authenticate with dbt Cloud.",
    )  # noqa: E501
    job_id: int = Field(
        default=..., description="The id of the dbt Cloud job.", title="Job ID"
    )
    timeout_seconds: int = Field(
        default=900,
        description="The number of seconds to wait for the job to complete.",
    )
    interval_seconds: int = Field(
        default=10,
        description="The number of seconds to wait between polling for job completion.",
    )
    trigger_job_run_options: TriggerJobRunOptions = Field(
        default_factory=TriggerJobRunOptions,
        description="The options to use when triggering a job run.",
    )

    @sync_compatible
    async def get_job(self, order_by: Optional[str] = None) -> Dict[str, Any]:
        """
        Retrieve information about a dbt Cloud job.

        Args:
            order_by: The field to order the results by.

        Returns:
            The job data.
        """
        try:
            async with self.dbt_cloud_credentials.get_administrative_client() as client:
                response = await client.get_job(
                    job_id=self.job_id,
                    order_by=order_by,
                )
        except HTTPStatusError as ex:
            raise DbtCloudGetJobFailed(extract_user_message(ex)) from ex
        return response.json()["data"]

    @sync_compatible
    async def trigger(
        self, trigger_job_run_options: Optional[TriggerJobRunOptions] = None
    ) -> DbtCloudJobRun:
        """
        Triggers a dbt Cloud job.

        Returns:
            A representation of the dbt Cloud job run.
        """
        try:
            trigger_job_run_options = (
                trigger_job_run_options or self.trigger_job_run_options
            )
            async with self.dbt_cloud_credentials.get_administrative_client() as client:
                response = await client.trigger_job_run(
                    job_id=self.job_id, options=trigger_job_run_options
                )
        except HTTPStatusError as ex:
            raise DbtCloudJobRunTriggerFailed(extract_user_message(ex)) from ex

        run_data = response.json()["data"]
        run_id = run_data.get("id")
        run = DbtCloudJobRun(
            dbt_cloud_job=self,
            run_id=run_id,
        )
        self.logger.info(
            f"dbt Cloud job {self.job_id} run {run_id} successfully triggered. "
            f"You can view the status of this run at "
            f"https://{self.dbt_cloud_credentials.domain}/#/accounts/"
            f"{self.dbt_cloud_credentials.account_id}/projects/"
            f"{run_data['project_id']}/runs/{run_id}/"
        )
        return run


@flow
async def run_dbt_cloud_job(
    dbt_cloud_job: DbtCloudJob,
    targeted_retries: int = 3,
) -> Dict[str, Any]:
    """
    Flow that triggers and waits for a dbt Cloud job run, retrying a
    subset of failed nodes if necessary.

    Args:
        dbt_cloud_job: Block that holds the information and
            methods to interact with a dbt Cloud job.
        targeted_retries: The number of times to retry failed steps.

    Examples:
        ```python
        from prefect import flow
        from prefect_dbt.cloud import DbtCloudCredentials, DbtCloudJob
        from prefect_dbt.cloud.jobs import run_dbt_cloud_job

        @flow
        def run_dbt_cloud_job_flow():
            dbt_cloud_credentials = DbtCloudCredentials.load("dbt-token")
            dbt_cloud_job = DbtCloudJob(
                dbt_cloud_credentials=dbt_cloud_credentials, job_id=154217
            )
            return run_dbt_cloud_job(dbt_cloud_job=dbt_cloud_job)

        run_dbt_cloud_job_flow()
        ```
    """
    logger = get_run_logger()

    run = await task(dbt_cloud_job.trigger.aio)(dbt_cloud_job)

    # Always try waiting for completion at least once
    try:
        await task(run.wait_for_completion.aio)(run)
        result = await task(run.fetch_result.aio)(run)
        return result
    except DbtCloudJobRunFailed:
        if targeted_retries <= 0:
            raise DbtCloudJobRunFailed(
                f"dbt Cloud job {run.run_id} failed after {targeted_retries} retries."
            )

        # Continue with retries if targeted_retries > 0
        remaining_retries = targeted_retries
        while remaining_retries > 0:
            logger.info(
                f"Retrying job run with ID: {run.run_id} {remaining_retries} more times"
            )
            run = await task(run.retry_failed_steps.aio)(run)
            remaining_retries -= 1
            try:
                await task(run.wait_for_completion.aio)(run)
                result = await task(run.fetch_result.aio)(run)
                return result
            except DbtCloudJobRunFailed:
                if remaining_retries <= 0:
                    break

        raise DbtCloudJobRunFailed(
            f"dbt Cloud job {run.run_id} failed after {targeted_retries} retries."
        )



"""Module containing models for Snowflake configs"""

from typing import Any, Dict, Optional

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


from pydantic import Field

from prefect_dbt.cli.configs.base import BaseTargetConfigs, MissingExtrasRequireError

try:
    from prefect_snowflake.database import SnowflakeConnector
except ModuleNotFoundError as e:
    raise MissingExtrasRequireError("Snowflake") from e


class SnowflakeTargetConfigs(BaseTargetConfigs):
    """
    Target configs contain credentials and
    settings, specific to Snowflake.
    To find valid keys, head to the [Snowflake Profile](
    https://docs.getdbt.com/reference/warehouse-profiles/snowflake-profile)
    page.

    Attributes:
        connector: The connector to use.

    Examples:
        Load stored SnowflakeTargetConfigs:
        ```python
        from prefect_dbt.cli.configs import SnowflakeTargetConfigs

        snowflake_target_configs = SnowflakeTargetConfigs.load("BLOCK_NAME")
        ```

        Instantiate SnowflakeTargetConfigs.
        ```python
        from prefect_dbt.cli.configs import SnowflakeTargetConfigs
        from prefect_snowflake.credentials import SnowflakeCredentials
        from prefect_snowflake.database import SnowflakeConnector

        credentials = SnowflakeCredentials(
            user="user",
            password="password",
            account="account.region.aws",
            role="role",
        )
        connector = SnowflakeConnector(
            schema="public",
            database="database",
            warehouse="warehouse",
            credentials=credentials,
        )
        target_configs = SnowflakeTargetConfigs(
            connector=connector,
            extras={"retry_on_database_errors": True},
        )
        ```
    """

    _block_type_name = "dbt CLI Snowflake Target Configs"
    _logo_url = "https://images.ctfassets.net/gm98wzqotmnx/5zE9lxfzBHjw3tnEup4wWL/9a001902ed43a84c6c96d23b24622e19/dbt-bit_tm.png?h=250"  # noqa
    _documentation_url = "https://docs.prefect.io/integrations/prefect-dbt"  # noqa

    type: Literal["snowflake"] = Field(
        default="snowflake", description="The type of the target configs."
    )
    schema_: Optional[str] = Field(
        default=None,
        alias="schema",
        description="The schema to use for the target configs.",
    )
    connector: SnowflakeConnector = Field(
        default=..., description="The connector to use."
    )

    def get_configs(self) -> Dict[str, Any]:
        """
        Returns the dbt configs specific to Snowflake profile.

        Returns:
            A configs JSON.
        """
        all_configs_json = super().get_configs()

        # decouple prefect-snowflake from prefect-dbt
        # by mapping all the keys dbt snowflake accepts
        # https://docs.getdbt.com/reference/warehouse-setups/snowflake-setup
        rename_keys = {
            # dbt
            "type": "type",
            "schema": "schema",
            "threads": "threads",
            # general
            "account": "account",
            "user": "user",
            "role": "role",
            "database": "database",
            "warehouse": "warehouse",
            # user and password
            "password": "password",
            # duo mfa / sso
            "authenticator": "authenticator",
            # key pair
            "private_key": "private_key",
            "private_key_path": "private_key_path",
            "private_key_passphrase": "private_key_passphrase",
            # optional
            "client_session_keep_alive": "client_session_keep_alive",
            "query_tag": "query_tag",
            "connect_retries": "connect_retries",
            "connect_timeout": "connect_timeout",
            "retry_on_database_errors": "retry_on_database_errors",
            "retry_all": "retry_all",
        }
        configs_json = {}
        extras = self.extras or {}
        for key in all_configs_json.keys():
            if key not in rename_keys and key not in extras:
                # skip invalid keys, like fetch_size + poll_frequency_s
                continue
            # rename key to something dbt profile expects
            dbt_key = rename_keys.get(key) or key
            value = all_configs_json[key]
            if key == "private_key":
                #  SnowflakeCredentials stores private_key as SecretBytes
                value = value.decode()
            configs_json[dbt_key] = value
        return configs_json



import pytest
import respx
from httpx import Response
from prefect_dbt.cloud.utils import (
    DbtCloudAdministrativeApiCallFailed,
    call_dbt_cloud_administrative_api_endpoint,
)


class TestCallDbtCloudAdministrativeApiEndpoint:
    async def test_endpoint_returns_json(self, dbt_cloud_credentials):
        with respx.mock(using="httpx") as respx_mock:
            respx_mock.get(
                "https://cloud.getdbt.com/api/v2/accounts/123456789/projects/",
                headers={"Authorization": "Bearer my_api_key"},
            ).mock(
                return_value=Response(
                    200,
                    json={
                        "status": {
                            "code": 200,
                            "is_success": True,
                            "user_message": "Success!",
                            "developer_message": "",
                        },
                        "data": [],
                    },
                )
            )

            result = await call_dbt_cloud_administrative_api_endpoint.fn(
                dbt_cloud_credentials=dbt_cloud_credentials,
                path="/projects/",
                http_method="GET",
            )

            assert result == {
                "status": {
                    "code": 200,
                    "is_success": True,
                    "user_message": "Success!",
                    "developer_message": "",
                },
                "data": [],
            }

    async def test_endpoint_returns_text(self, dbt_cloud_credentials):
        with respx.mock(using="httpx") as respx_mock:
            respx_mock.get(
                "https://cloud.getdbt.com/api/v2/accounts/123456789/runs/12/artifacts/compiled/dbt_artifacts/models/dim_dbt__current_models.sql",  # noqa
                headers={"Authorization": "Bearer my_api_key"},
            ).mock(return_value=Response(200, text="Hi! I'm some SQL!"))

            result = await call_dbt_cloud_administrative_api_endpoint.fn(
                dbt_cloud_credentials=dbt_cloud_credentials,
                path="/runs/12/artifacts/compiled/dbt_artifacts/models/dim_dbt__current_models.sql",  # noqa
                http_method="GET",
            )

            assert result == "Hi! I'm some SQL!"

    async def test_failure(self, dbt_cloud_credentials):
        with respx.mock(using="httpx") as respx_mock:
            respx_mock.get(
                "https://cloud.getdbt.com/api/v2/accounts/123456789/projects/",
                headers={"Authorization": "Bearer my_api_key"},
            ).mock(return_value=Response(500, json={}))

            with pytest.raises(DbtCloudAdministrativeApiCallFailed):
                await call_dbt_cloud_administrative_api_endpoint.fn(
                    dbt_cloud_credentials=dbt_cloud_credentials,
                    path="/projects/",
                    http_method="GET",
                )



from pathlib import Path

from prefect_dbt.cli.configs import SnowflakeTargetConfigs
from prefect_snowflake.credentials import SnowflakeCredentials
from prefect_snowflake.database import SnowflakeConnector
from pydantic import SecretBytes, SecretStr


def test_snowflake_target_configs_get_configs():
    credentials = SnowflakeCredentials(
        account="account",
        user="user",
        password="password",
    )
    snowflake_connector = SnowflakeConnector(
        schema="schema",
        database="database",
        warehouse="warehouse",
        credentials=credentials,
    )
    configs = SnowflakeTargetConfigs(
        connector=snowflake_connector, extras={"retry_on_database_errors": True}
    )

    actual = configs.get_configs()
    expected = dict(
        account="account",
        user="user",
        password="password",
        type="snowflake",
        schema="schema",
        database="database",
        warehouse="warehouse",
        authenticator="snowflake",
        retry_on_database_errors=True,
        threads=4,
    )
    for k, v in actual.items():
        actual_v = (
            v.get_secret_value() if isinstance(v, (SecretBytes, SecretStr)) else v
        )
        expected_v = expected[k]
        assert actual_v == expected_v


def test_snowflake_target_configs_get_configs_private_key_path():
    credentials = SnowflakeCredentials(
        account="account",
        user="user",
        private_key_path=Path("path/to/key"),
    )
    snowflake_connector = SnowflakeConnector(
        schema="schema",
        database="database",
        warehouse="warehouse",
        credentials=credentials,
    )
    configs = SnowflakeTargetConfigs(
        connector=snowflake_connector, extras={"retry_on_database_errors": True}
    )

    actual = configs.get_configs()
    expected = dict(
        account="account",
        user="user",
        private_key_path=str(Path("path/to/key")),
        type="snowflake",
        schema="schema",
        database="database",
        warehouse="warehouse",
        authenticator="snowflake",
        retry_on_database_errors=True,
        threads=4,
    )
    for k, v in actual.items():
        actual_v = (
            v.get_secret_value() if isinstance(v, (SecretBytes, SecretStr)) else v
        )
        expected_v = expected[k]
        assert actual_v == expected_v


def test_snowflake_target_configs_allow_field_overrides():
    credentials = SnowflakeCredentials(
        user="user",
        account="account.region.aws",
        password="my_password",
        role="role",
    )

    connector = SnowflakeConnector(
        schema="public",
        database="database",
        warehouse="warehouse",
        credentials=credentials,
    )

    target_configs = SnowflakeTargetConfigs(
        connector=connector,
        schema="OVERRIDE",
        extras={"database": "my_database"},
        allow_field_overrides=True,
    )
    actual = target_configs.get_configs()
    expected = {
        "schema": "OVERRIDE",
        "type": "snowflake",
        "threads": 4,
        "account": "account.region.aws",
        "user": "user",
        "password": "my_password",
        "authenticator": "snowflake",
        "role": "role",
        "database": "my_database",
        "warehouse": "warehouse",
    }
    assert actual == expected


def test_snowflake_target_configs_get_configs_private_key():
    credentials = SnowflakeCredentials(
        account="account",
        user="user",
        private_key="some-private-key",
    )
    snowflake_connector = SnowflakeConnector(
        schema="schema",
        database="database",
        warehouse="warehouse",
        credentials=credentials,
    )
    configs = SnowflakeTargetConfigs(
        connector=snowflake_connector, extras={"retry_on_database_errors": True}
    )

    actual = configs.get_configs()
    expected = dict(
        account="account",
        user="user",
        private_key="some-private-key",
        type="snowflake",
        schema="schema",
        database="database",
        warehouse="warehouse",
        authenticator="snowflake",
        retry_on_database_errors=True,
        threads=4,
    )
    for k, v in actual.items():
        actual_v = (
            v.get_secret_value() if isinstance(v, (SecretBytes, SecretStr)) else v
        )
        expected_v = expected[k]
        assert actual_v == expected_v


def test_snowflake_target_configs_schema_from_connector_in_profile():
    """
    Regression test for issue #19069.

    When creating a DbtCliProfile from SnowflakeTargetConfigs with a connector,
    the schema from the connector should be included in the profile output,
    even when not explicitly passed to SnowflakeTargetConfigs.
    """
    from prefect_dbt.cli import DbtCliProfile

    credentials = SnowflakeCredentials(
        account="account.region.aws",
        user="user",
        password="password",
    )
    connector = SnowflakeConnector(
        schema="ANALYTICS_PROD",
        database="ANALYTICS",
        warehouse="FIVETRAN_WAREHOUSE",
        credentials=credentials,
    )
    # Note: NOT passing schema explicitly to SnowflakeTargetConfigs
    target_configs = SnowflakeTargetConfigs(connector=connector)
    profile = DbtCliProfile(
        name="dubclub", target="prod", target_configs=target_configs
    ).get_profile()

    # The schema should be present in the profile output
    assert "dubclub" in profile
    assert "outputs" in profile["dubclub"]
    assert "prod" in profile["dubclub"]["outputs"]

    prod_output = profile["dubclub"]["outputs"]["prod"]
    assert "schema" in prod_output, (
        f"schema missing from profile output. Available keys: {list(prod_output.keys())}"
    )
    assert prod_output["schema"] == "ANALYTICS_PROD"



from prefect_dbt.cli.configs import PostgresTargetConfigs
from prefect_sqlalchemy import (
    ConnectionComponents,
    SqlAlchemyConnector,
)
from pydantic import SecretStr


def test_postgres_target_configs_get_configs_for_sqlalchemy_connector():
    configs = PostgresTargetConfigs(
        credentials=SqlAlchemyConnector(
            connection_info=ConnectionComponents(
                driver="postgresql+psycopg2",
                database="postgres",
                username="prefect",
                password="prefect_password",
                host="host",
                port=8080,
                query={"a": "query"},
            ),
            kwarg_that_shouldnt_show_up=True,
        ),
        schema="schema",
        extras={"retries": 1},
    )
    actual = configs.get_configs()
    expected = {
        "type": "postgres",
        "schema": "schema",
        "threads": 4,
        "dbname": "postgres",
        "user": "prefect",
        "password": "prefect_password",
        "host": "host",
        "port": 8080,
        "retries": 1,
    }
    for k, v in actual.items():
        actual_v = v.get_secret_value() if isinstance(v, SecretStr) else v
        expected_v = expected[k]
        assert actual_v == expected_v
    assert hasattr(configs.credentials, "kwarg_that_shouldnt_show_up")
    assert "kwarg_that_shouldnt_show_up" not in actual.keys()
    assert hasattr(configs.credentials.connection_info, "query")
    assert "query" not in actual.keys()



"""
State for managing tasks across callbacks.
"""

import threading
from typing import Any, Optional, Union
from uuid import UUID

from prefect.client.schemas.objects import Flow, State
from prefect.context import hydrated_context
from prefect.logging.loggers import PrefectLogAdapter, get_logger
from prefect.task_engine import run_task_sync
from prefect.tasks import Task


class NodeTaskTracker:
    """
    Maintains the state, upstream dependencies, logger, and results
    for all the tasks called by the execution of each node in the dbt project.
    """

    def __init__(self):
        self._tasks: dict[str, Task[Any, Any]] = {}
        self._task_results: dict[str, Any] = {}
        self._node_status: dict[str, dict[str, Any]] = {}
        self._node_complete: dict[str, bool] = {}
        self._node_dependencies: dict[str, list[str]] = {}
        self._node_events: dict[str, threading.Event] = {}
        self._task_run_ids: dict[str, UUID] = {}
        self._task_run_names: dict[str, str] = {}

    def start_task(self, node_id: str, task: Task[Any, Any]) -> None:
        """Start a task for a node."""
        self._tasks[node_id] = task
        self._node_complete[node_id] = False
        self._node_events[node_id] = threading.Event()

    def get_task_logger(
        self,
        node_id: str,
        flow_run: Optional[dict[str, Any]] = None,
        flow: Optional[Flow] = None,
        **kwargs: Any,
    ) -> PrefectLogAdapter:
        """Get the logger for a task."""
        logger = PrefectLogAdapter(
            get_logger("prefect.task_runs"),
            extra={
                **{
                    "task_run_id": self.get_task_run_id(node_id),
                    "flow_run_id": str(flow_run.get("id")) if flow_run else "<unknown>",
                    "task_run_name": self.get_task_run_name(node_id),
                    "task_name": "execute_dbt_node",
                    "flow_run_name": flow_run.get("name") if flow_run else "<unknown>",
                    "flow_name": flow.name if flow else "<unknown>",
                },
                **kwargs,
            },
        )

        return logger

    def set_node_status(
        self, node_id: str, event_data: dict[str, Any], event_message: str
    ) -> None:
        """Set the status for a node."""
        self._node_status[node_id] = {
            "event_data": event_data,
            "event_message": event_message,
        }
        # Mark node as complete when status is set
        self._node_complete[node_id] = True
        # Signal the event to wake up any waiting threads
        if node_id in self._node_events:
            self._node_events[node_id].set()

    def get_node_status(self, node_id: str) -> Union[dict[str, Any], None]:
        """Get the status for a node."""
        return self._node_status.get(node_id)

    def is_node_complete(self, node_id: str) -> bool:
        """Check if a node is complete."""
        return self._node_complete.get(node_id, False)

    def wait_for_node_completion(
        self, node_id: str, timeout: Union[float, None] = None
    ) -> bool:
        """Wait for a node to complete using threading.Event.

        Args:
            node_id: The ID of the node to wait for
            timeout: Maximum time to wait in seconds. None means wait indefinitely.

        Returns:
            True if the node completed, False if timeout occurred
        """
        if node_id not in self._node_events:
            # If no event exists, the node might already be complete
            return self.is_node_complete(node_id)

        return self._node_events[node_id].wait(timeout=timeout)

    def set_task_result(self, node_id: str, result: Any) -> None:
        """Set the result for a task."""
        self._task_results[node_id] = result

    def get_task_result(self, node_id: str) -> Union[Any, None]:
        """Get the result for a task."""
        return self._task_results.get(node_id)

    def set_node_dependencies(self, node_id: str, dependencies: list[str]) -> None:
        """Set the dependencies for a node."""
        self._node_dependencies[node_id] = dependencies

    def get_node_dependencies(self, node_id: str) -> list[str]:
        """Get the dependencies for a node."""
        return self._node_dependencies.get(node_id, [])

    def set_task_run_id(self, node_id: str, task_run_id: UUID) -> None:
        """Set the task run ID for a node."""
        self._task_run_ids[node_id] = task_run_id

    def get_task_run_id(self, node_id: str) -> Union[UUID, None]:
        """Get the task run ID for a node."""
        return self._task_run_ids.get(node_id)

    def set_task_run_name(self, node_id: str, task_run_name: str) -> None:
        """Set the task run name for a node."""
        self._task_run_names[node_id] = task_run_name

    def get_task_run_name(self, node_id: str) -> Union[str, None]:
        """Get the task run name for a node."""
        return self._task_run_names.get(node_id)

    def run_task_in_thread(
        self,
        node_id: str,
        task: Task[Any, Any],
        task_run_id: UUID,
        parameters: dict[str, Any],
        context: dict[str, Any],
    ) -> None:
        """Run a task in a separate thread."""

        def run_task():
            with hydrated_context(context):
                states: list[State] = []
                dependencies = self.get_node_dependencies(node_id)
                for dep_id in dependencies:
                    state = self.get_task_result(dep_id)
                    if state:
                        states.append(state)

                state = run_task_sync(
                    task,
                    task_run_id=task_run_id,
                    parameters=parameters,
                    wait_for=states,
                    context=context,
                    return_type="state",
                )

                # Wait for the task to complete
                if state:
                    self.set_task_result(node_id, state)
                else:
                    self.set_task_result(node_id, None)

        thread = threading.Thread(target=run_task)
        thread.daemon = True
        thread.start()



import pytest
from prefect_dbt import (
    DbtCloudCredentials,
    DbtCloudJob,
    PrefectDbtSettings,
)


def test_direct_imports():
    """Test that directly imported objects are available."""
    assert PrefectDbtSettings is not None
    assert DbtCloudCredentials is not None
    assert DbtCloudJob is not None


def test_lazy_import_dbt_cli_profile():
    """Test that DbtCliProfile can be lazily imported."""
    from prefect_dbt import DbtCliProfile

    assert DbtCliProfile is not None


def test_lazy_import_global_configs():
    """Test that GlobalConfigs can be lazily imported."""
    from prefect_dbt import GlobalConfigs

    assert GlobalConfigs is not None


def test_lazy_import_target_configs():
    """Test that TargetConfigs can be lazily imported."""
    from prefect_dbt import TargetConfigs

    assert TargetConfigs is not None


def test_lazy_import_database_configs():
    """Test that database-specific configs can be lazily imported."""
    from prefect_dbt import (
        BigQueryTargetConfigs,
        PostgresTargetConfigs,
        SnowflakeTargetConfigs,
    )

    assert PostgresTargetConfigs is not None
    assert BigQueryTargetConfigs is not None
    assert SnowflakeTargetConfigs is not None


def test_invalid_attribute():
    """Test that accessing an invalid attribute raises ImportError."""
    with pytest.raises(ImportError) as exc_info:
        from prefect_dbt import NonExistentAttribute  # noqa: F401

    assert "cannot import name 'NonExistentAttribute' from 'prefect_dbt'" in str(
        exc_info.value
    )


def test_all_exports():
    """Test that __all__ contains expected exports."""
    import prefect_dbt

    assert hasattr(prefect_dbt, "__version__")
    assert hasattr(prefect_dbt, "PrefectDbtSettings")
    assert hasattr(prefect_dbt, "PrefectDbtRunner")
    assert hasattr(prefect_dbt, "DbtCloudCredentials")
    assert hasattr(prefect_dbt, "DbtCloudJob")
